# Functions to control the nodes' operations.

import glob
import logging
import os
import shutil
import time
from collections import namedtuple

from ZeekControl import cmdresult, config, cron, events, execute, install, util
from ZeekControl import node as node_mod


# Waits for the nodes' Zeek processes to reach the given status.
# Build the Zeek parameters for the given node. Include
# script for live operation if live is true.
def _make_zeek_params(node, live):
    args = []

    if live and node.interface:
        try:
            # Interface name needs quotes so that shell doesn't interpret any
            # potential metacharacters in the name.
            args += ["-i", f"'{node.interface}'"]
        except AttributeError:
            pass

        if config.Config.savetraces:
            args += ["-w", "trace.pcap"]

    args += ["-U", ".status"]
    args += ["-p", "zeekctl"]

    if live:
        args += ["-p", "zeekctl-live"]

    if node_mod.is_standalone(node):
        args += ["-p", "standalone"]

    for prefix in config.Config.prefixes.split(":"):
        args += ["-p", f"{prefix}"]

    args += ["-p", f"{node.name}"]

    # The order of loaded scripts is as follows:
    # 1) SitePolicyScripts (local.zeek by default) gives a common set of loaded
    #    scripts for all nodes.
    # 2) The common configuration of zeekctl is loaded via the zeekctl package.
    # 3) The distribution's default settings for node configuration are loaded
    #    from either the cluster framework or standalone scripts. At this point
    #    anything in the distribution's default per-node is overridable and any
    #    identifiers in local.zeek are able to be used (e.g. in defining
    #    a notice policy).
    # 4) Autogenerated zeekctl scripts are loaded, which may contain
    #    settings that override the previously loaded scripts.
    #    (e.g. see Log::default_rotation_interval)
    args += config.Config.sitepolicyscripts.split()
    args += ["zeekctl"]
    if node_mod.is_standalone(node):
        args += ["zeekctl/standalone"]
    else:
        args += ["base/frameworks/cluster"]

    args += ["zeekctl/auto"]

    if getattr(node, "aux_scripts", None):
        args += [node.aux_scripts]

    if config.Config.zeekargs:
        # Some args in zeekargs might contain spaces, so we cannot split it.
        args += [config.Config.zeekargs]

    return args


# Build the environment variables for the given node.
def _make_env_params(node, returnlist=False):
    envs = []
    if not node_mod.is_standalone(node):
        envs.append(f"CLUSTER_NODE={node.name}")

    envs += [f"{key}={val}" for (key, val) in sorted(node.env_vars.items())]

    if returnlist:
        envlist = [("-v", i) for i in envs]
        return [j for i in envlist for j in i]

    return " ".join(envs)


def fmttime(t):
    return time.strftime(config.Config.timefmt, time.localtime(float(t)))


class Controller:
    def __init__(self, config, ui, executor, pluginregistry):
        self.config = config
        self.ui = ui
        self.executor = executor
        self.pluginregistry = pluginregistry

        # Create zeekctl-config.sh file so that shell script helpers have
        # current config values.
        install.make_zeekctl_config_sh(ui)

    def start(self, nodes):
        results = cmdresult.CmdResult()

        loggers, manager, proxies, workers = node_mod.separate_types(nodes)

        for n in nodes:
            n.setExpectRunning(True)

        # Start nodes. Do it in the order loggers, manager, proxies, workers.
        if loggers:
            self._start_nodes(loggers, results)

            if not results.ok:
                for n in manager + proxies + workers:
                    results.set_node_fail(n)
                return results

        if manager:
            self._start_nodes(manager, results)

            if not results.ok:
                for n in proxies + workers:
                    results.set_node_fail(n)
                return results

        if proxies:
            self._start_nodes(proxies, results)

            if not results.ok:
                for n in workers:
                    results.set_node_fail(n)
                return results

        if workers:
            self._start_nodes(workers, results)

        return results

    # Starts the given nodes.
    def _start_nodes(self, nodes, results):
        self.ui.info(f"starting {node_mod.nodes_describe(nodes)} ...")

        filtered = []
        # Ignore nodes which are still running.
        for node, isrunning in self._isrunning(nodes):
            if not isrunning:
                filtered += [node]

        nodes = filtered

        # Generate crash report for any crashed nodes.
        crashed = [node for node in nodes if node.hasCrashed()]
        if crashed:
            self.ui.info(
                "creating crash report for previously crashed nodes: {}".format(
                    ", ".join([n.name for n in crashed])
                )
            )
            self._make_crash_reports(crashed)

        # Make working directories.
        dirs = [(node, node.cwd()) for node in nodes]
        nodes = []
        for node, success, output in self.executor.mkdirs(dirs):
            if success:
                nodes += [node]
            else:
                self.ui.error(f"cannot create working directory for {node.name}")
                results.set_node_fail(node)

        # Start Zeek process.
        cmds = []
        for node in nodes:
            envs = []
            pin_cpu = node.pin_cpus

            # If this node isn't using CPU pinning, then use a placeholder value
            if pin_cpu == "":
                pin_cpu = -1

            envs = _make_env_params(node, True)
            cmds += [
                (
                    node,
                    "start",
                    envs + [node.cwd(), str(pin_cpu)] + _make_zeek_params(node, True),
                )
            ]

        nodes = []
        # Note: the shell is used to interpret the command because zeekargs
        # might contain quoted arguments.
        for node, success, output in self.executor.run_helper(cmds, shell=True):
            if success:
                if not output:
                    self.ui.error(f"failed to get PID of {node.name}")
                    results.set_node_fail(node)
                    continue

                pidstr = output.splitlines()[0]
                try:
                    pid = int(pidstr)
                except ValueError:
                    self.ui.error(f"invalid PID for {node.name}: {pidstr}")
                    results.set_node_fail(node)
                    continue

                nodes += [node]
                node.setPID(pid)
            else:
                self.ui.error(f'cannot start {node.name}; check output of "diag"')
                results.set_node_fail(node)
                if output:
                    self.ui.error(output)

        # Check whether processes did indeed start up.
        hanging = []
        running = []

        for node, success in self._waitforzeeks(nodes, "RUNNING", 3, True):
            if success:
                running += [node]
            else:
                hanging += [node]

        # It can happen that Zeek hangs in DNS lookups at startup
        # which can take a while. At this point we already know
        # that the process has been started (_waitforzeeks ensures that).
        # If by now there is not a TERMINATED status, we assume that it
        # is doing fine and will move on to RUNNING once DNS is done.
        for node, success in self._waitforzeeks(hanging, "TERMINATED", 0, False):
            if success:
                self.ui.error(
                    f'{node.name} terminated immediately after starting; check output with "diag"'
                )
                node.clearPID()
                results.set_node_fail(node)
            else:
                self.ui.info(f"({node.name} still initializing)")
                running += [node]

        for node in running:
            self._log_action(node, "started")
            results.set_node_success(node)

        return results

    def _isrunning(self, nodes, setcrashed=True):
        results = []
        cmds = []

        for node in nodes:
            pid = node.getPID()
            if not pid:
                results += [(node, False)]
                continue

            cmds += [(node, "check-pid", [str(pid)])]

        for node, success, output in self.executor.run_helper(cmds):
            # If we cannot run the helper script, then we ignore this node
            # because the process might actually be running but we can't tell.
            if not success:
                self.ui.error(f"failed to run check-pid on node {node.name}")
                continue

            running = output.strip() == "running"

            results += [(node, running)]

            if not running:
                if setcrashed:
                    # Grmpf. It crashed.
                    node.clearPID()
                    node.setCrashed()

        return results

    def _waitforzeeks(self, nodes, status, timeout, ensurerunning):
        # If ensurerunning is true, process must still be running.
        if ensurerunning:
            running = self._isrunning(nodes)
        else:
            running = [(node, True) for node in nodes]

        results = []

        # Determine set of nodes still to check.
        todo = {}
        for node, isrunning in running:
            if isrunning:
                todo[node.name] = node
            else:
                results += [(node, False)]

        while True:
            # Determine whether process is still running. We need to do this
            # before we get the state to avoid a race condition.

            nodelist = sorted(todo.values(), key=node_mod.sortnode)
            running = self._isrunning(nodelist, setcrashed=False)

            # Check nodes' .status file
            cmds = []
            for node in nodelist:
                cmds += [(node, "first-line", [f"{node.cwd()}/.status"])]

            for node, success, output in self.executor.run_helper(cmds):
                if not success or not output:
                    continue

                fields = output.split()
                if len(fields) == 2:
                    if status in fields[0]:
                        # Status reached. Cool.
                        del todo[node.name]
                        results += [(node, True)]
                else:
                    # Something's wrong. We give up on that node.
                    del todo[node.name]
                    results += [(node, False)]

            for node, isrunning in running:
                if node.name in todo and not isrunning:
                    # Alright, a dead node's status will not change anymore.
                    del todo[node.name]
                    results += [(node, False)]

            if not todo:
                # All done.
                break

            # Wait a bit before we start over.
            time.sleep(1)

            # Timeout reached?
            timeout -= 1
            if timeout <= 0:
                break

            logging.debug("Waiting for %d node(s)...", len(todo))

        for node in todo.values():
            # These did time-out.
            results += [(node, False)]

        if todo:
            logging.debug("Timeout while waiting for %d node(s)", len(todo))

        return results

    def _log_action(self, node, action):
        if not self.config.statslogenable:
            return
        t = time.time()
        with open(self.config.statslog, "a") as out:
            out.write(f"{t} {node} action {action}\n")

    # Do a "post-terminate crash" for the given nodes.
    def _make_crash_reports(self, nodes):
        for n in nodes:
            self.pluginregistry.zeekProcessDied(n)

        msg_header_backtrace = "If you want to help us debug this problem, then please forward\nthis mail to reports@zeek.org\n"

        msg_header_no_backtrace = "This crash report does not include a backtrace.  In order for crash reports\nto be useful when Zeek crashes, a backtrace is needed.\n"

        postterminate = os.path.join(self.config.scriptsdir, "post-terminate")
        cmds = [
            (node, postterminate, [node.type, node.cwd(), "crash"]) for node in nodes
        ]

        for node, success, output in self.executor.run_cmds(cmds):
            if success:
                crashreport = output

                # Note: here it is assumed that the crash-diag script outputs
                # this string only when there's a backtrace.
                has_backtrace = "Core file: " in crashreport

                if has_backtrace:
                    msg = msg_header_backtrace + crashreport
                else:
                    msg = msg_header_no_backtrace + crashreport

                msuccess, moutput = self._sendmail(
                    f"Crash report from {node.name}", msg
                )
                if not msuccess:
                    self.ui.error(
                        f"error occurred while trying to send mail: {moutput}"
                    )
            else:
                self.ui.error(
                    f"error running post-terminate for {node.name}:\n{output}"
                )

            node.clearCrashed()

    def _sendmail(self, subject, body):
        if not self.config.sendmail:
            return True, ""

        cmd = "{} '{}'".format(
            os.path.join(self.config.scriptsdir, "send-mail"), subject
        )
        return execute.run_localcmd(cmd, inputtext=body)

    # Stop Zeek processes on nodes.
    def stop(self, nodes):
        results = cmdresult.CmdResult()

        loggers, manager, proxies, workers = node_mod.separate_types(nodes)

        for n in nodes:
            n.setExpectRunning(False)

        # Stop nodes. Do it in the order workers, proxies, manager, loggers
        # (the reverse of "start").
        if workers:
            self._stop_nodes(workers, results)

            if not results.ok:
                for n in proxies + manager + loggers:
                    results.set_node_fail(n)
                return results

        if proxies:
            self._stop_nodes(proxies, results)

            if not results.ok:
                for n in manager + loggers:
                    results.set_node_fail(n)
                return results

        if manager:
            self._stop_nodes(manager, results)

            if not results.ok:
                for n in loggers:
                    results.set_node_fail(n)
                return results

        if loggers:
            self._stop_nodes(loggers, results)

        return results

    def _stop_nodes(self, nodes, results):
        self.ui.info(f"stopping {node_mod.nodes_describe(nodes)} ...")

        running = []

        # Check which nodes are still running.
        for node, isrunning in self._isrunning(nodes):
            if isrunning:
                running += [node]
            else:
                results.set_node_success(node)

        # Generate crash report for any crashed nodes.
        crashed = [node for node in nodes if node.hasCrashed()]
        if crashed:
            self.ui.info(
                "creating crash report for previously crashed nodes: {}".format(
                    ", ".join([n.name for n in crashed])
                )
            )
            self._make_crash_reports(crashed)

        # Helper function to stop nodes with given signal.
        def stop(nodes, signal):
            cmds = []
            for node in nodes:
                cmds += [(node, "stop", [str(node.getPID()), str(signal)])]

            return self.executor.run_helper(cmds)

        # Stop nodes.
        for node, success, output in stop(running, 15):
            if not success:
                # Give up on this node.  Most likely either we cannot connect
                # to the host, or we don't have permission to kill the process.
                self.ui.error(f"unable to stop {node.name}: {output}")
                results.set_node_fail(node)
                running.remove(node)

        if running:
            time.sleep(1)

        # Check whether they terminated.
        terminated = []
        kill = []
        for node, success in self._waitforzeeks(
            running, "TERMINATED", self.config.stoptimeout, False
        ):
            if not success:
                # Check whether it crashed during shutdown ...
                result = self._isrunning([node])
                for node, isrunning in result:
                    if isrunning:
                        self.ui.info(f"{node.name} did not terminate ... killing ...")
                        kill += [node]
                    else:
                        # crashed flag is set by _isrunning().
                        self.ui.info(f"{node.name} crashed during shutdown")

        if kill:
            # Kill those which did not terminate gracefully.
            stop(kill, 9)
            # Give them a bit to disappear.
            time.sleep(5)

        # Check which are still running. We check all nodes to be on the safe
        # side and give them a bit more time to finally disappear.
        timeout = 10

        todo = {}
        for node in running:
            todo[node.name] = node

        while True:
            nodelist = sorted(todo.values(), key=node_mod.sortnode)
            running = self._isrunning(nodelist, setcrashed=False)

            for node, isrunning in running:
                if node.name in todo and not isrunning:
                    # Alright, it's gone.
                    del todo[node.name]
                    terminated += [node]
                    results.set_node_success(node)

            if not todo:
                # All done.
                break

            # Wait a bit before we start over.

            if timeout <= 0:
                break

            time.sleep(1)
            timeout -= 1

        for node in todo.values():
            results.set_node_fail(node)

        # Do post-terminate cleanup for those which terminated gracefully.
        cleanup = [node for node in terminated if not node.hasCrashed()]

        cmds = []
        postterminate = os.path.join(self.config.scriptsdir, "post-terminate")
        for node in cleanup:
            crashflag = "killed" if node in kill else ""

            cmds += [(node, postterminate, [node.type, node.cwd(), crashflag])]

        for node, success, output in self.executor.run_cmds(cmds):
            if success:
                self._log_action(node, "stopped")
            else:
                self.ui.error(
                    f"error running post-terminate for {node.name}:\n{output}"
                )
                self._log_action(node, "stopped (failed)")

            node.clearPID()
            node.clearCrashed()

        return results

    # Output status summary for nodes.
    def status(self, nodes):
        results = cmdresult.CmdResult()

        showall = self.config.statuscmdshowall

        if showall:
            self.ui.info("Getting process status ...")

        nodestatus = self._isrunning(nodes)
        running = []

        cmds = []
        for node, isrunning in nodestatus:
            if isrunning:
                running += [node]
                cmds += [
                    (
                        node,
                        "first-line",
                        [f"{node.cwd()}/.status", f"{node.cwd()}/.startup"],
                    )
                ]

        statuses = {}
        startups = {}
        for n, success, output in self.executor.run_helper(cmds):
            out = output.splitlines()
            try:
                val = out[0].split()[0].lower() if (success and out[0]) else "???"
            except IndexError:
                val = "???"

            statuses[n.name] = val

            try:
                val = fmttime(out[1]) if (success and out[1]) else "???"
            except (IndexError, ValueError):
                val = "???"

            startups[n.name] = val

        if showall:
            self.ui.info("Getting peer status ...")
            peers = {}
            nodes = [n for n in running if statuses[n.name] == "running"]
            for node, success, args in self._query_peerstatus(nodes):
                if success and args:
                    peers[node.name] = []
                    for f in args[0].split():
                        if not f.startswith("peer="):
                            continue
                        # Get everything after the '=' character.
                        val = f[5:]
                        if val:
                            peers[node.name] += [val]

        for node, isrunning in nodestatus:
            node_info = {
                "name": node.name,
                "type": node.type,
                "host": node.host,
                "status": "stopped",
                "pid": None,
                "started": None,
            }
            if showall:
                node_info["peers"] = None

            if isrunning:
                node_info["status"] = statuses[node.name]
            elif node.hasCrashed():
                node_info["status"] = "crashed"

            if isrunning:
                node_info["pid"] = node.getPID()

                if showall:
                    if node.name in peers:
                        node_info["peers"] = len(peers[node.name])
                    else:
                        node_info["peers"] = "???"

                node_info["started"] = startups[node.name]

            results.set_node_data(node, True, node_info)

        return results

    # Check the configuration for nodes without installing first.
    def check(self, nodes):
        return self._check_config(nodes, False, False)

    # Print the loaded_scripts.log for either the installed scripts
    # (if "check" is false), or the original scripts (if "check" is true).
    def scripts(self, nodes, check):
        return self._check_config(nodes, not check, True)

    def _check_config(self, nodes, installed, list_scripts):
        results = cmdresult.CmdResult()

        nodetmpdirs = [
            (node, os.path.join(self.config.tmpdir, f"check-config-{node.name}"))
            for node in nodes
        ]

        nodes = []
        for node, cwd in nodetmpdirs:
            if os.path.isdir(cwd):
                try:
                    shutil.rmtree(cwd)
                except OSError as err:
                    self.ui.error(f"cannot remove directory {cwd}: {err}")
                    results.ok = False
                    return results

            try:
                os.makedirs(cwd)
            except OSError as err:
                self.ui.error(f"cannot create temporary directory: {err}")
                results.ok = False
                return results

            nodes += [(node, cwd)]

        cmds = []
        for node, cwd in nodes:
            env = _make_env_params(node)

            installed_policies = "1" if installed else "0"
            print_scripts = "1" if list_scripts else "0"

            if not install.make_layout(cwd, self.ui, True):
                results.ok = False
                return results
            if not install.make_local_networks(cwd, self.ui):
                results.ok = False
                return results

            if not install.make_zeekctl_config_policy(
                cwd, self.ui, self.pluginregistry
            ):
                results.ok = False
                return results

            cmd = os.path.join(
                self.config.scriptsdir, "check-config"
            ) + " {} {} {} {}".format(
                installed_policies,
                print_scripts,
                cwd,
                " ".join(_make_zeek_params(node, False)),
            )
            cmd += " zeekctl/check"

            cmds += [((node, cwd), cmd, env, None)]

        for (node, cwd), success, output in execute.run_localcmds(cmds):
            results.set_node_output(node, success, output)
            try:
                shutil.rmtree(cwd)
            except OSError:
                # Don't bother reporting an error now.
                pass

        return results

    def _query_peerstatus(self, nodes):
        running = self._isrunning(nodes)

        eventlist = []
        for node, isrunning in running:
            if isrunning:
                eventlist += [
                    (
                        node,
                        "Control::peer_status_request",
                        [],
                        "Control::peer_status_response",
                    )
                ]

        return events.send_events_parallel(eventlist, config.Config.controltopic)

    def execute_cmd(self, nodes, cmd):
        results = cmdresult.CmdResult()

        for node, success, out in self.executor.run_shell_cmds(
            [(n, cmd) for n in nodes]
        ):
            results.set_node_output(node, success, out)

        return results

    # Clean up the working directory for nodes (flushes state).
    # If cleantmp is true, also wipes ${tmpdir}; this is done
    # even when the node is still running.
    def cleanup(self, nodes, cleantmp=False):
        # Given a set of node names "orig" and command results "res", add
        # all node names to "orig" that have a failed result in "res".
        def addfailed(orig, res):
            for node, status, output in res:
                # if status is Fail, then add the node name
                if not status:
                    orig.add(node.name)

            return orig

        results = cmdresult.CmdResult()

        result = self._isrunning(nodes)
        running = [node for (node, on) in result if on]
        notrunning = [node for (node, on) in result if not on]

        for node in running:
            self.ui.info(f"   {node} is still running, not cleaning work directory")

        results1 = self.executor.rmdirs([(n, n.cwd()) for n in notrunning])
        results2 = self.executor.mkdirs([(n, n.cwd()) for n in notrunning])
        failed = set()
        failed = addfailed(failed, results1)
        failed = addfailed(failed, results2)

        for node in notrunning:
            node.clearCrashed()

        if cleantmp:
            self.ui.info(f"cleaning {self.config.tmpdir} ...")
            results3 = self.executor.rmdirs(
                [(n, self.config.tmpdir) for n in running + notrunning]
            )
            results4 = self.executor.mkdirs(
                [(n, self.config.tmpdir) for n in running + notrunning]
            )
            failed = addfailed(failed, results3)
            failed = addfailed(failed, results4)

        for node in nodes:
            if node.name in failed:
                results.set_node_fail(node)
            else:
                results.set_node_success(node)

        return results

    # Report diagnostics for nodes (e.g., stderr output).
    def diag(self, nodes):
        results = cmdresult.CmdResult()

        crashdiag = os.path.join(self.config.scriptsdir, "crash-diag")
        cmds = [(node, crashdiag, [node.cwd()]) for node in nodes]

        for node, success, output in self.executor.run_cmds(cmds):
            if not success:
                errmsgs = f"error running crash-diag for {node.name}\n"
                errmsgs += output
                results.set_node_output(node, False, errmsgs)
                continue

            results.set_node_output(node, True, output)

        return results

    def capstats(self, nodes, interval):
        results = cmdresult.CmdResult()

        if not self.config.capstatspath:
            results.set_node_data(
                nodes[0],
                False,
                {
                    "output": 'Error: cannot run capstats because zeekctl option "capstatspath" is not defined'
                },
            )
            return results

        for node, netif, success, vals in self.get_capstats_output(nodes, interval):
            if not success:
                vals = {"output": vals}
            results.set_node_data(node, success, vals)

        if not results.nodes:
            results.ok = False

        return results

    # Gather capstats from interfaces.
    #
    # Returns a list of tuples of the form (node, netif, success, vals)
    # where 'netif' is the network interface name used by capstats on
    # the 'node', and 'success' is a boolean indicating whether or not
    # we were able to get the data; in case there's no error, 'vals' maps
    # tags to their values (otherwise, 'vals' is an error message).
    #
    # Tags are those as returned by capstats on the command-line.
    #
    # If there is more than one node, then the results will also contain
    # one "pseudo-node" of the name "$total" with the sum of all individual
    # values.
    def get_capstats_output(self, nodes, interval):
        results = []

        # Construct a list of (node, interface) tuples, one tuple for each
        # unique (host, interface) pair.
        nodenetifs = []
        hosts = {}
        for node in nodes:
            if not node.interface:
                continue

            netif = self._capstats_interface(node)
            if not netif:
                continue

            if hosts.setdefault((node.addr, netif), node) == node:
                nodenetifs.append((node, netif))

        capstats = self.config.capstatspath
        cmds = [
            (node, capstats, ["-I", str(interval), "-n", "1", "-i", interface])
            for (node, interface) in nodenetifs
        ]

        outputs = self.executor.run_cmds(cmds)

        totals = {}

        for node, success, output in outputs:
            netif = self._capstats_interface(node)

            if output:
                # Grab the first output line, because we might log this to
                # stats.log later.
                outputline = output.splitlines()[0]

            if not success:
                if output:
                    results += [
                        (
                            node,
                            netif,
                            False,
                            f"{node.name}: capstats failed ({outputline})",
                        )
                    ]
                else:
                    results += [
                        (node, netif, False, f"{node.name}: cannot execute capstats")
                    ]
                continue

            if not output:
                results += [(node, netif, False, f"{node.name}: no capstats output")]
                continue

            fields = outputline.split()[1:]

            if not fields:
                results += [
                    (
                        node,
                        netif,
                        False,
                        f"{node.name}: unexpected capstats output: {outputline}",
                    )
                ]
                continue

            vals = {}

            try:
                for field in fields:
                    key, val = field.split("=")
                    val = float(val)
                    vals[key] = val

                    if key in totals:
                        totals[key] += val
                    else:
                        totals[key] = val

            except ValueError:
                results += [
                    (
                        node,
                        netif,
                        False,
                        f"{node.name}: unexpected capstats output: {outputline}",
                    )
                ]
                continue

            results += [(node, netif, True, vals)]

        # Add pseudo-node for totals when there is more than one result
        if len(results) > 1:
            results += [(node_mod.Node(self.config, "$total"), None, True, totals)]

        return results

    # Convert a Zeek network interface name to one that capstats can use.
    def _capstats_interface(self, node):
        netif = node.interface

        if netif.startswith("dnacl") and netif.count("@") == 1:
            # PF_RING+DNA with pfdnacluster_master is being used
            # (e.g. interface name "dnacluster:21" gets changed to
            # "dnacluster:21@1" by the zeekctl pf_ring plugin)
            netif = netif.split("@", 1)[0]

        elif "::" in netif:
            # Interface name has packet source prefix (e.g. "af_packet::eth0"),
            # so don't try to run capstats on this interface unless it is
            # af_packet since we know that works.

            if netif.startswith("af_packet"):
                netif = netif.split("::")[1]
            else:
                netif = None

        return netif

    # Gets disk space on all volumes relevant to zeekctl installation.
    # Returns a list of the form:  [ (host, diskinfo), ...]
    # where diskinfo is a list of the form DiskInfo named tuple objects (fs,
    # total, used, avail, percent) or ["FAIL", <error message>] if an error
    # is encountered.
    def df(self, nodes):
        results = cmdresult.CmdResult()

        DiskInfo = namedtuple(
            "DiskInfo", ("fs", "total", "used", "available", "percent")
        )
        dirs = (
            "logdir",
            "bindir",
            "helperdir",
            "cfgdir",
            "spooldir",
            "policydir",
            "libdir",
            "libdir64",
            "tmpdir",
            "staticdir",
            "scriptsdir",
        )

        df = {}
        for node in nodes:
            df[node.name] = {}

        cmds = []
        for node in nodes:
            for key in dirs:
                if key == "logdir" and not (
                    node_mod.is_logger(node)
                    or node_mod.is_manager(node)
                    or node_mod.is_standalone(node)
                ):
                    # Don't need to check this on nodes that don't write logs.
                    continue

                path = self.config.config[key]

                if key == "libdir" or key == "libdir64":
                    if not os.path.exists(path):
                        continue

                cmds += [(node, "df", [path])]

        for node, success, output in self.executor.run_helper(cmds):
            if success:
                fields = output.split()
                if len(fields) != 4:
                    df[node.name]["FAIL"] = "wrong number of fields from df helper"
                    continue

                fs = fields[0]
                # Ignore NFS mounted volumes.
                if not fs.startswith("/") and ":" in fs:
                    continue

                try:
                    total = float(fields[1])
                    used = float(fields[2])
                    avail = float(fields[3])
                except ValueError as err:
                    df[node.name]["FAIL"] = f"bad output from df helper: {err}"
                    continue

                perc = used * 100.0 / (used + avail)
                df[node.name][fs] = DiskInfo(fs, total, used, avail, perc)
            else:
                df[node.name]["FAIL"] = output if output else "no output"

        for node in nodes:
            success = "FAIL" not in df[node.name]
            results.set_node_data(node, success, df[node.name])

        return results

    # Returns a list of tuples of the form (node, error, vals) where 'error' is
    # an error message string, or None if there was no error.  'vals' is a
    # dict which maps tags to their values.  Tags are "pid", "vsize",
    # "rss", "cpu", and "cmd".
    def get_top_output(self, nodes):
        results = []

        running = self._isrunning(nodes)

        # Get all the PIDs first.

        pids = {}

        for node, isrunning in running:
            if isrunning:
                pids[node.name] = node.getPID()
            else:
                results += [(node, "not running", {})]
                continue

        if not pids:
            return results

        cmds = []
        hosts = {}

        # Now run top once per host.
        for node in nodes:  # Do the loop again to keep the order.
            if node.name not in pids:
                continue

            if node.host in hosts:
                continue

            hosts[node.host] = 1

            cmds += [(node, "top", [])]

        if not cmds:
            return results

        res = {}
        for node, success, output in self.executor.run_helper(cmds):
            res[node.host] = success, output

        # Gather results for all the nodes that are running
        for node in nodes:
            if node.name not in pids:
                continue

            success, output = res[node.host]

            if not success:
                # The error msg gets written to stats.log, so we only want
                # the first line.
                errmsg = output.splitlines()[0] if output else ""
                results += [(node, f"top failed: {errmsg}", {})]
                continue

            if not output:
                results += [(node, "no output from top", {})]
                continue

            # Get the zeek process info, which is a list of fields from
            # the "top" helper.
            procinfo = []
            try:
                for line in output.splitlines():
                    if int(line.split()[0]) == pids[node.name]:
                        procinfo = line.split()
                        break
            except (IndexError, ValueError) as err:
                results += [(node, f"bad output from top: {err}", {})]
                continue

            if not procinfo:
                # It's possible that the process is no longer there.
                results += [(node, "not running", {})]
                continue

            vals = {}

            try:
                pid = int(procinfo[0])
                vals["pid"] = pid
                vals["vsize"] = int(
                    float(procinfo[1])
                )  # May be something like 2.17684e+9
                vals["rss"] = int(float(procinfo[2]))
                vals["cpu"] = procinfo[3]
                vals["cmd"] = " ".join(procinfo[4:])
            except (IndexError, ValueError) as err:
                results += [(node, f"unexpected top output: {err}", {})]
                continue

            results += [(node, None, vals)]

        return results

    # Produce a top-like output for node's processes.
    def top(self, nodes):
        results = cmdresult.CmdResult()

        for node, error, vals in self.get_top_output(nodes):
            top_info = {
                "name": node.name,
                "type": node.type,
                "host": node.host,
                "pid": None,
                "vsize": None,
                "rss": None,
                "cpu": None,
                "cmd": None,
                "error": None,
            }
            if error:
                top_info["error"] = error
                results.set_node_data(node, False, {"procs": top_info})
                continue

            top_info2 = top_info.copy()
            top_info2.update(vals)

            results.set_node_data(node, True, {"procs": top_info2})

        return results

    def print_id(self, nodes, id):
        results = cmdresult.CmdResult()
        running = self._isrunning(nodes)

        eventlist = []
        for node, isrunning in running:
            if isrunning:
                eventlist += [
                    (
                        node,
                        "Control::id_value_request",
                        [id],
                        "Control::id_value_response",
                    )
                ]

        if not eventlist:
            results.set_node_output(nodes[0], False, "no running instances of Zeek")
            return results

        for node, success, args in events.send_events_parallel(
            eventlist, config.Config.controltopic
        ):
            if success:
                out = "\n".join(args)
            else:
                out = args
            results.set_node_output(node, success, out)

        return results

    def _query_netstats(self, nodes):
        running = self._isrunning(nodes)

        eventlist = []
        for node, isrunning in running:
            if isrunning:
                eventlist += [
                    (
                        node,
                        "Control::net_stats_request",
                        [],
                        "Control::net_stats_response",
                    )
                ]

        return events.send_events_parallel(eventlist, config.Config.controltopic)

    def peerstatus(self, nodes):
        results = cmdresult.CmdResult()
        for node, success, args in self._query_peerstatus(nodes):
            if success:
                if args:
                    out = args[0]
                else:
                    out = ""
            else:
                out = args
            results.set_node_output(node, success, out)

        if not results.nodes:
            results.set_node_output(nodes[0], False, "no running instances of Zeek")

        return results

    def netstats(self, nodes):
        results = cmdresult.CmdResult()
        for node, success, args in self._query_netstats(nodes):
            if success:
                if args:
                    out = args[0].strip()
                else:
                    out = ""
            else:
                out = args
            results.set_node_output(node, success, out)

        if not results.nodes:
            results.set_node_output(nodes[0], False, "no running instances of Zeek")

        return results

    def process(self, trace, zeek_options, zeek_scripts):
        results = cmdresult.CmdResult()

        if not os.path.isfile(trace):
            self.ui.error(f"trace file not found: {trace}")
            results.ok = False
            return results

        if self.config.standalone:
            node = self.config.nodes()[0]
        else:
            node = self.config.workers()[0]

        cwd = os.path.join(self.config.tmpdir, "testing")

        if os.path.isdir(cwd):
            try:
                shutil.rmtree(cwd)
            except OSError as err:
                self.ui.error(f"cannot remove directory: {err}")
                results.ok = False
                return results

        try:
            os.makedirs(cwd)
        except OSError as err:
            self.ui.error(f"cannot create directory: {err}")
            results.ok = False
            return results

        env = _make_env_params(node)

        zeek_args = " ".join(zeek_options + _make_zeek_params(node, False))
        zeek_args += " zeekctl/process-trace"

        if zeek_scripts:
            zeek_args += " " + " ".join(zeek_scripts)

        cmd = (
            os.path.join(self.config.scriptsdir, "run-zeek-on-trace")
            + f" {0} {cwd} {trace} {zeek_args}"
        )

        self.ui.info(cmd)

        success, output = execute.run_localcmd(cmd, env=env)

        if not success:
            results.ok = False

        self.ui.info(output)
        self.ui.info(f"### Zeek output in {cwd}")

        return results

    def install(self, local_only):
        results = cmdresult.CmdResult()

        try:
            self.config.record_zeek_version()
        except config.ConfigurationError as err:
            self.ui.error(f"{err}")
            results.ok = False
            return results

        manager = self.config.manager()

        # Delete previously installed policy files to not mix things up.
        policies = [
            self.config.policydirsiteinstall,
            self.config.policydirsiteinstallauto,
        ]

        for dirpath in policies:
            if os.path.isdir(dirpath):
                self.ui.info(f"removing old policies in {dirpath} ...")
                try:
                    shutil.rmtree(dirpath)
                except OSError as err:
                    self.ui.error(f"failed to remove directory {dirpath}: {err}")
                    results.ok = False
                    return results

        self.ui.info("creating policy directories ...")
        for dirpath in policies:
            try:
                os.makedirs(dirpath)
            except OSError as err:
                self.ui.error(f"failed to create directory: {err}")
                results.ok = False
                return results

        # Install local site policy.

        if self.config.sitepolicypath:
            self.ui.info("installing site policies ...")
            dst = self.config.policydirsiteinstall
            for dir in self.config.sitepolicypath.split(":"):
                dirpath = self.config.subst(dir)
                for pathname in glob.glob(os.path.join(dirpath, "*")):
                    if not execute.install(pathname, dst, self.ui):
                        results.ok = False
                        return results

        if not install.make_layout(self.config.policydirsiteinstallauto, self.ui):
            results.ok = False
            return results

        self.ui.info("generating local-networks.zeek ...")
        if not install.make_local_networks(
            self.config.policydirsiteinstallauto, self.ui
        ):
            results.ok = False
            return results

        self.ui.info("generating zeekctl-config.zeek ...")
        if not install.make_zeekctl_config_policy(
            self.config.policydirsiteinstallauto, self.ui, self.pluginregistry
        ):
            results.ok = False
            return results

        loggers = self.config.loggers()
        if loggers:
            # Just use the first logger that is defined.
            node_cwd = loggers[0].cwd()
        else:
            node_cwd = manager.cwd()

        current = self.config.subst(os.path.join(self.config.logdir, "current"))
        try:
            util.force_symlink(node_cwd, current)
        except OSError as err:
            results.ok = False
            self.ui.error(f"failed to update symlink '{current}': {err}")
            return results

        self.ui.info("generating zeekctl-config.sh ...")
        if not install.make_zeekctl_config_sh(self.ui):
            results.ok = False
            return results

        if local_only:
            return results

        # Make sure we install each remote host only once.
        nodes = self.config.hosts(exclude_local=True)

        # If there are no remote hosts, then we're done.
        if not nodes:
            # Save current configuration state.
            self.config.update_cfg_hash()
            return results

        # Sync to clients.
        self.ui.info("updating nodes ...")

        dirs = []

        if not self.config.havenfs:
            # Non-NFS, need to explicitly synchronize.
            syncs = install.get_syncs()
        else:
            # NFS. We only need to take care of the spool/log directories.

            # We need this only on the manager.
            dirs.append((manager, self.config.logdir))

            syncs = install.get_nfssyncs()

        syncs = [
            (dir, mirror)
            for (dir, mirror, optional) in syncs
            if not optional or os.path.exists(self.config.subst(dir))
        ]

        createdirs = [self.config.subst(dir) for (dir, mirror) in syncs if not mirror]
        for n in nodes:
            for dir in createdirs:
                dirs.append((n, dir))

        for node, success, output in self.executor.mkdirs(dirs):
            if not success:
                self.ui.error(f"cannot create a directory on node {node.name}")
                if output:
                    self.ui.error(output)
                results.ok = False
                return results

        paths = [self.config.subst(dir) for (dir, mirror) in syncs if mirror]
        if not execute.sync(nodes, paths, self.ui):
            results.ok = False
            return results

        # Save current configuration state.
        self.config.update_cfg_hash()

        return results

    # Triggers all activity which is to be done regularly via cron.
    def cron(self, watch):
        if not self.config.cronenabled:
            logging.debug("cron is disabled")
            return

        # Check if "zeekctl install" has been run.
        if not self.config.is_zeekctl_installed():
            # Don't output anything here, otherwise the cron job may generate
            # emails before the user has a chance to do "zeekctl install".
            return

        cronui = cron.CronUI()
        tasks = cron.CronTasks(
            cronui, self.config, self, self.executor, self.pluginregistry
        )

        cronui.buffer_output()

        if watch:
            # Check if node state matches expected state, and start/stop if
            # necessary.
            startlist = []
            stoplist = []
            for node, isrunning in self._isrunning(self.config.nodes()):
                expectrunning = node.getExpectRunning()

                if not isrunning and expectrunning:
                    startlist.append(node)
                elif isrunning and not expectrunning:
                    stoplist.append(node)

            if startlist:
                self.start(startlist)
            if stoplist:
                self.stop(stoplist)

        # Check for dead hosts.
        tasks.check_hosts()

        # Generate statistics.
        tasks.log_stats(5)

        # Check available disk space.
        tasks.check_disk_space()

        # Expire old log files.
        tasks.expire_logs()

        # Expire old crash directories.
        tasks.expire_crash()

        # Update the HTTP stats directory.
        tasks.update_http_stats()

        # Run external command if we have one.
        tasks.run_cron_cmd()

        # Mail potential output.
        output = cronui.get_buffered_output()
        if output:
            success, out = self._sendmail("cron: " + output.splitlines()[0], output)
            if not success:
                self.ui.error(f"zeekctl cron failed to send mail: {out}")
                self.ui.info(f"Output of zeekctl cron:\n{output}")

        logging.debug("cron done")



import ZeekControl.config
import ZeekControl.plugin


class ClusterBackendCheck(ZeekControl.plugin.Plugin):
    def __init__(self):
        super().__init__(apiversion=1)

    def name(self):
        return "cluster_backend_check"

    def pluginVersion(self):
        return 1

    def init(self):
        return True

    def cmd_install_pre(self):
        """
        Ensure that the internal ClusterTopicSeparator is set. If it isn't,
        then the selected ClusterBackend plugin doesn't work right.
        """
        backend = self.getGlobalOption("ClusterBackend")
        topic_sep = self.getGlobalOption("ClusterTopicSeparator")

        if not topic_sep or len(topic_sep) == 0:
            self.error(
                f"internal ClusterTopicSeparator '{topic_sep}' invalid - is cluster backend '{backend}' working properly?"
            )
            return False

        return True



# The ZeekControl interactive shell.

import logging
import os
import sys

from ZeekControl import cmdresult, config, control, execute, lock, pluginreg, version
from ZeekControl import node as node_mod
from ZeekControl.exceptions import InvalidNodeError, LockError, RuntimeEnvironmentError


class TermUI:
    def info(self, txt):
        print(txt)

    def error(self, txt):
        print(txt, file=sys.stderr)

    warn = error


class NullHandler(logging.Handler):
    def emit(self, record):
        pass


def expose(func):
    func.api_exposed = True
    return func


def lock_required(func):
    def wrapper(self, *args, **kwargs):
        self.lock()
        try:
            return func(self, *args, **kwargs)
        finally:
            self.unlock()

    wrapper.lock_required = True
    return wrapper


def lock_required_silent(func):
    def wrapper(self, *args, **kwargs):
        self.lock(showwait=False)
        try:
            return func(self, *args, **kwargs)
        finally:
            self.unlock()

    wrapper.lock_required = True
    return wrapper


def check_config(func):
    def wrapper(self, *args, **kwargs):
        if config.Config.is_cfg_changed():
            self.ui.warn('Configuration has changed. Run the "deploy" command.')
        return func(self, *args, **kwargs)

    return wrapper


class ZeekCtl:
    def __init__(
        self,
        basedir=version.ZEEKBASE,
        libdir=version.LIBDIR,
        libdirinternal=version.LIBDIRINTERNAL,
        cfgfile=version.CFGFILE,
        zeekscriptdir=version.ZEEKSCRIPTDIR,
        ui=TermUI(),
        state=None,
    ):
        self.ui = ui
        self.zeekbase = basedir
        self.libdir = libdir
        self.libdirinternal = libdirinternal

        self.config = config.Configuration(
            self.zeekbase,
            self.libdir,
            self.libdirinternal,
            cfgfile,
            zeekscriptdir,
            self.ui,
            state,
        )

        # Remove all log handlers (set by any previous calls to logging.*)
        logging.getLogger().handlers = []

        if self.config.debug:
            # Add a log handler that logs to a file.
            try:
                logging.basicConfig(
                    filename=self.config.debuglog,
                    format="%(asctime)s [%(module)s] %(message)s",
                    datefmt=self.config.timefmt,
                    level=logging.DEBUG,
                )
            except OSError as err:
                raise RuntimeEnvironmentError(
                    f"{err}\nCheck if the user running ZeekControl has write access to the debug log file."
                )
        else:
            # Add a log handler that does nothing.
            h = NullHandler()
            logging.getLogger().addHandler(h)

        self.executor = execute.Executor(self.config)
        self.plugins = pluginreg.PluginRegistry()
        self.setup()
        self.controller = control.Controller(
            self.config, self.ui, self.executor, self.plugins
        )

    def setup(self):
        plugindirs = self.config.sitepluginpath.split(":")
        plugindirs.append(self.config.plugindir)
        plugindirs.append(self.config.pluginzeekdir)

        for pdir in plugindirs:
            if pdir:
                self.plugins.addDir(pdir)

        self.plugins.loadPlugins(self.ui, self.executor)
        self.plugins.initPluginOptions()
        self.plugins.addNodeKeys()
        self.config.initPostPlugins()
        self.plugins.initPlugins(self.ui)
        self.plugins.initPluginCmds()
        os.chdir(self.config.zeekbase)
        if self.config.get_state("cronenabled") is None:
            self.config.set_state("cronenabled", True)

    def reload_cfg(self):
        self.config.reload_cfg()

        if self.config.debug:
            if isinstance(logging.getLogger().handlers[0], NullHandler):
                # Remove the null handler and configure logging to a file.
                logging.getLogger().handlers = []
                logging.basicConfig(
                    filename=self.config.debuglog,
                    format="%(asctime)s [%(module)s] %(message)s",
                    datefmt=self.config.timefmt,
                    level=logging.DEBUG,
                )

            # Re-enable all log levels.
            logging.disable(logging.NOTSET)
        else:
            # Disable logging to all log levels that we use.
            logging.disable(logging.CRITICAL)

        self.executor.finish()
        self.plugins.initPluginOptions()
        self.config.initPostPlugins()
        self.plugins.initPlugins(self.ui)
        self.plugins.initPluginCmds()

    def finish(self):
        self.executor.finish()
        self.plugins.finishPlugins()

    def warn_zeekctl_install(self):
        self.config.warn_zeekctl_install()

    # Turns node name arguments into a list of nodes.  If "get_hosts" is True,
    # then only one node per host is chosen.  If "get_types" is True, then
    # only one node per node type (manager, proxy, etc.) is chosen.
    def node_args(self, args=None, get_hosts=False, get_types=False):
        nodes = []

        if args:
            for arg in args.split():
                nodelist = self.config.nodes(arg)
                if not nodelist:
                    raise InvalidNodeError(f"unknown node '{arg}'")

                nodes += nodelist

            # Remove duplicate nodes
            newlist = list(set(nodes))
            if len(newlist) != len(nodes):
                nodes = newlist
        else:
            # Get all nodes.
            nodes = self.config.nodes()

        # Sort the list so that it doesn't depend on initial order of arguments
        nodes.sort(key=node_mod.sortnode)

        if get_hosts:
            hosts = {}
            hostnodes = []
            for node in nodes:
                if node.host not in hosts:
                    hosts[node.host] = 1
                    hostnodes.append(node)

            nodes = hostnodes

        if get_types:
            types = {}
            typenodes = []
            for node in nodes:
                if node.type not in types:
                    types[node.type] = 1
                    typenodes.append(node)

            nodes = typenodes

        return nodes

    def lock(self, showwait=True):
        lockstatus = lock.lock(self.ui, showwait)
        if not lockstatus:
            raise LockError("Unable to get lock")

        self.config.read_state()

    def unlock(self):
        lock.unlock(self.ui)

    def node_names(self):
        return [n.name for n in self.config.nodes()]

    def node_groups(self):
        return node_mod.node_groups()

    @expose
    @check_config
    def nodes(self):
        results = cmdresult.CmdResult()

        if self.plugins.cmdPre("nodes"):
            for n in self.config.nodes():
                results.set_node_data(n, True, n.to_dict())
        else:
            results.ok = False

        self.plugins.cmdPost("nodes")

        return results

    @expose
    @check_config
    def get_config(self):
        results = cmdresult.CmdResult()

        if self.plugins.cmdPre("config"):
            results.keyval = self.config.options()
        else:
            results.ok = False

        self.plugins.cmdPost("config")
        return results

    @expose
    @check_config
    @lock_required
    def install(self, local=False):
        if self.plugins.cmdPre("install"):
            results = self.controller.install(local)
        else:
            results = cmdresult.CmdResult(ok=False)

        self.plugins.cmdPost("install")
        return results

    @expose
    @check_config
    @lock_required
    def start(self, node_list=None):
        nodes = self.node_args(node_list)

        nodes = self.plugins.cmdPreWithNodes("start", nodes)
        results = self.controller.start(nodes)
        self.plugins.cmdPostWithResults("start", results.get_node_data())

        return results

    @expose
    @check_config
    @lock_required
    def stop(self, node_list=None):
        nodes = self.node_args(node_list)

        nodes = self.plugins.cmdPreWithNodes("stop", nodes)
        results = self.controller.stop(nodes)
        self.plugins.cmdPostWithResults("stop", results.get_node_data())

        return results

    @expose
    @check_config
    @lock_required
    def restart(self, clean=False, node_list=None):
        nodes = self.node_args(node_list)

        nodes = self.plugins.cmdPreWithNodes("restart", nodes, clean)

        self.ui.info("stopping ...")
        results = self.stop(node_list)
        if not results.ok:
            return results

        if clean:
            self.ui.info("cleaning up ...")
            results = self.cleanup(node_list=node_list)
            if not results.ok:
                return results

            self.ui.info("checking configurations ...")
            results = self.check(node_list)
            if not results.ok:
                return results

            self.ui.info("installing ...")
            results = self.install()
            if not results.ok:
                return results

        self.ui.info("starting ...")
        results = self.start(node_list)

        self.plugins.cmdPostWithNodes("restart", nodes)
        return results

    @expose
    @lock_required
    def deploy(self):
        if not self.plugins.cmdPre("deploy"):
            results = cmdresult.CmdResult(ok=False)
            return results

        if self.config.is_cfg_changed():
            self.ui.info("Reloading zeekctl configuration ...")
            self.reload_cfg()

        self.ui.info("checking configurations ...")
        results = self.check(check_node_types=True)
        if not results.ok:
            for node, success, output in results.get_node_output():
                if not success:
                    self.ui.info(f"{node} scripts failed.")
                    self.ui.info(output)

            return results

        self.ui.info("installing ...")
        results = self.install()
        if not results.ok:
            return results

        self.ui.info("stopping ...")
        results = self.stop()
        if not results.ok:
            return results

        self.ui.info("starting ...")
        results = self.start()

        self.plugins.cmdPost("deploy")
        return results

    @expose
    @check_config
    @lock_required
    def status(self, node_list=None):
        nodes = self.node_args(node_list)

        nodes = self.plugins.cmdPreWithNodes("status", nodes)
        results = self.controller.status(nodes)
        self.plugins.cmdPostWithNodes("status", nodes)
        return results

    @expose
    @lock_required
    def top(self, node_list=None):
        nodes = self.node_args(node_list)

        nodes = self.plugins.cmdPreWithNodes("top", nodes)
        results = self.controller.top(nodes)
        self.plugins.cmdPostWithNodes("top", nodes)

        return results

    @expose
    @check_config
    @lock_required
    def diag(self, node_list=None):
        nodes = self.node_args(node_list)

        nodes = self.plugins.cmdPreWithNodes("diag", nodes)
        results = self.controller.diag(nodes)
        self.plugins.cmdPostWithNodes("diag", nodes)

        return results

    @expose
    @lock_required_silent
    def cron(self, watch=True):
        if self.plugins.cmdPre("cron", "", watch):
            self.controller.cron(watch)
        self.plugins.cmdPost("cron", "", watch)

        return True

    @expose
    @check_config
    @lock_required
    def cronenabled(self):
        results = False
        if self.plugins.cmdPre("cron", "?", False):
            results = self.config.cronenabled
        self.plugins.cmdPost("cron", "?", False)
        return results

    @expose
    @check_config
    @lock_required
    def setcronenabled(self, enable=True):
        if enable:
            if self.plugins.cmdPre("cron", "enable", False):
                self.config.set_state("cronenabled", True)
                self.ui.info("cron enabled")
            self.plugins.cmdPost("cron", "enable", False)
        else:
            if self.plugins.cmdPre("cron", "disable", False):
                self.config.set_state("cronenabled", False)
                self.ui.info("cron disabled")
            self.plugins.cmdPost("cron", "disable", False)

        return True

    @expose
    @check_config
    @lock_required
    def check(self, node_list=None, check_node_types=False):
        nodes = self.node_args(node_list, get_types=check_node_types)

        nodes = self.plugins.cmdPreWithNodes("check", nodes)
        results = self.controller.check(nodes)
        self.plugins.cmdPostWithResults("check", results.get_node_data())

        return results

    @expose
    @check_config
    @lock_required
    def cleanup(self, cleantmp=False, node_list=None):
        nodes = self.node_args(node_list)

        nodes = self.plugins.cmdPreWithNodes("cleanup", nodes, cleantmp)
        results = self.controller.cleanup(nodes, cleantmp)
        self.plugins.cmdPostWithNodes("cleanup", nodes, cleantmp)

        return results

    @expose
    @check_config
    @lock_required
    def capstats(self, interval=10, node_list=None):
        nodes = self.node_args(node_list)
        nodes = self.plugins.cmdPreWithNodes("capstats", nodes, interval)
        results = self.controller.capstats(nodes, interval)
        self.plugins.cmdPostWithNodes("capstats", nodes, interval)

        return results

    @expose
    @check_config
    @lock_required
    def df(self, node_list=None):
        nodes = self.node_args(node_list, get_hosts=True)
        nodes = self.plugins.cmdPreWithNodes("df", nodes)
        results = self.controller.df(nodes)
        self.plugins.cmdPostWithNodes("df", nodes)

        return results

    @expose
    @check_config
    @lock_required
    def print_id(self, id, node_list=None):
        nodes = self.node_args(node_list)
        nodes = self.plugins.cmdPreWithNodes("print", nodes, id)
        results = self.controller.print_id(nodes, id)
        self.plugins.cmdPostWithNodes("print", nodes, id)

        return results

    @expose
    @check_config
    @lock_required
    def peerstatus(self, node_list=None):
        nodes = self.node_args(node_list)
        nodes = self.plugins.cmdPreWithNodes("peerstatus", nodes)
        results = self.controller.peerstatus(nodes)
        self.plugins.cmdPostWithNodes("peerstatus", nodes)

        return results

    @expose
    @check_config
    @lock_required
    def netstats(self, node_list=None):
        if not node_list:
            node_list = None
            if not self.config.standalone:
                node_list = node_mod.worker_group()

        nodes = self.node_args(node_list)
        nodes = self.plugins.cmdPreWithNodes("netstats", nodes)
        results = self.controller.netstats(nodes)
        self.plugins.cmdPostWithNodes("netstats", nodes)

        return results

    @expose
    @check_config
    def execute(self, cmd):
        nodes = self.node_args(get_hosts=True)

        if self.plugins.cmdPre("exec", cmd):
            results = self.controller.execute_cmd(nodes, cmd)
        else:
            results = cmdresult.CmdResult(ok=False)

        self.plugins.cmdPost("exec", cmd)

        return results

    @expose
    @check_config
    @lock_required
    def scripts(self, check=False, node_list=None):
        nodes = self.node_args(node_list)

        nodes = self.plugins.cmdPreWithNodes("scripts", nodes, check)
        results = self.controller.scripts(nodes, check)
        self.plugins.cmdPostWithNodes("scripts", nodes, check)

        return results

    @expose
    @check_config
    @lock_required
    def process(self, trace, options, scripts):
        if self.plugins.cmdPre("process", trace, options, scripts):
            results = self.controller.process(trace, options, scripts)
        else:
            results = cmdresult.CmdResult(ok=False)

        self.plugins.cmdPost("process", trace, options, scripts, results.ok)

        return results

    @expose
    @check_config
    @lock_required
    def plugincmd(self, cmd, args):
        return self.plugins.runCustomCommand(cmd, args, self.ui)



# Curses functions

import atexit
import curses
import signal

_Stdscr = None


def _finishCurses():
    curses.nocbreak()
    curses.echo()
    curses.endwin()


def _initCurses():
    global _Stdscr
    atexit.register(_finishCurses)
    _Stdscr = curses.initscr()


def enterCurses():
    if not _Stdscr:
        _initCurses()

    curses.cbreak()
    curses.noecho()
    _Stdscr.nodelay(1)

    signal.signal(signal.SIGWINCH, signal.SIG_IGN)


def leaveCurses():
    curses.reset_shell_mode()
    signal.signal(signal.SIGWINCH, signal.SIG_DFL)


# Check non-blockingly for a key press and returns it, or return None if no
# key is found. enter/leaveCurses must surround the getc() call.
def getCh():
    ch = _Stdscr.getch()

    if ch < 0:
        return None

    return chr(ch)


def clearScreen():
    if not _Stdscr:
        _initCurses()

    _Stdscr.clear()


def printLines(lines):
    y = 0
    for line in lines:
        try:
            _Stdscr.insnstr(y, 0, line, len(line))
        except:
            pass
        y += 1

    try:
        _Stdscr.insnstr(y, 0, "", 0)
    except:
        pass



# This plugin sets necessary environment variables to run Zeek with
# myricom load balancing.

import ZeekControl.plugin


class LBMyricom(ZeekControl.plugin.Plugin):
    def __init__(self):
        super().__init__(apiversion=1)

    def name(self):
        return "lb_myricom"

    def pluginVersion(self):
        return 1

    def init(self):
        useplugin = False

        for nn in self.nodes():
            if nn.type != "worker" or nn.lb_method != "myricom":
                continue

            useplugin = True

            # Apply environment variables, but do not override values from
            # the node.cfg or zeekctl.cfg files.
            nn.env_vars.setdefault("SNF_NUM_RINGS", nn.lb_procs)
            nn.env_vars.setdefault("SNF_FLAGS", "0x101")

        return useplugin



# Tasks which are to be done on a regular basis from cron.
import io
import os
import shutil
import time

from ZeekControl import execute
from ZeekControl import node as node_mod


class CronUI:
    def __init__(self):
        self.buffer = None

    def info(self, txt):
        if self.buffer:
            self.buffer.write(f"{txt}\n")
        else:
            print(txt)

    error = info
    warn = info

    def buffer_output(self):
        self.buffer = io.StringIO()

    def get_buffered_output(self):
        buf = self.buffer.getvalue()
        self.buffer.close()
        self.buffer = None
        return buf


class CronTasks:
    def __init__(self, ui, config, controller, executor, pluginregistry):
        self.ui = ui
        self.config = config
        self.controller = controller
        self.executor = executor
        self.pluginregistry = pluginregistry

    def log_stats(self, interval):
        if not self.config.statslogenable:
            return

        nodes = self.config.nodes()
        top = self.controller.get_top_output(nodes)

        have_capstats = self.config.capstatspath
        capstats = []

        if have_capstats:
            capstats = self.controller.get_capstats_output(nodes, interval)

        t = time.time()

        try:
            with open(self.config.statslog, "a") as out:
                for node, error, vals in top:
                    if not error:
                        for val, key in sorted(vals.items()):
                            out.write(f"{t} {node} parent {val} {key}\n")
                    else:
                        out.write(f"{t} {node} error error {error}\n")

                for node, netif, success, vals in capstats:
                    if not success:
                        out.write(f"{t} {node} error error {vals}\n")
                        continue

                    for key, val in sorted(vals.items()):
                        out.write(f"{t} {node} interface {key} {val}\n")

                        if key == "pkts" and str(node) != "$total":
                            # Report if we don't see packets on an interface.
                            tag = f"lastpkts-{node.name}"

                            last = self.config.get_state(tag, default=-1.0)

                            if self.config.mailreceivingpackets:
                                if val == 0.0 and last != 0.0:
                                    self.ui.info(
                                        f"{node.host} is not seeing any packets on interface {netif}"
                                    )

                                if val != 0.0 and last == 0.0:
                                    self.ui.info(
                                        f"{node.host} is seeing packets again on interface {netif}"
                                    )

                            self.config.set_state(tag, val)

        except OSError as err:
            self.ui.error(f"failed to append to file: {err}")
            return

    def check_disk_space(self):
        minspace = self.config.mindiskspace
        if minspace == 0:
            return

        results = self.controller.df(self.config.hosts())
        for node, _, dfs in results.get_node_data():
            host = node.host

            for key, df in dfs.items():
                if key == "FAIL":
                    # A failure here is normally caused by a host that is down,
                    # so we don't need to output the error message.
                    continue

                fs = df.fs
                perc = df.percent
                key = "disk-space-{}{}".format(host, fs.replace("/", "-"))

                if perc > 100 - minspace:
                    last = self.config.get_state(key, default=-1)
                    if last > 100 - minspace:
                        # Already reported.
                        continue

                    self.ui.warn(f"Disk space low on {host}:{fs} - {perc:.1f}% used.")

                self.config.set_state(key, perc)

    def expire_logs(self):
        if (
            self.config.logexpireminutes == 0
            and self.config.statslogexpireinterval == 0
        ):
            return

        if self.config.standalone:
            success, output = execute.run_localcmd(
                os.path.join(self.config.scriptsdir, "expire-logs")
            )

            if not success:
                self.ui.error(f"expire-logs failed\n{output}")
        else:
            nodes = self.config.hosts(tag=node_mod.logger_group())

            if not nodes:
                nodes = self.config.hosts(tag=node_mod.manager_group())

            expirelogs = os.path.join(self.config.scriptsdir, "expire-logs")
            cmds = [(node, expirelogs, []) for node in nodes]

            for node, success, output in self.executor.run_cmds(cmds):
                if not success:
                    self.ui.error(f"expire-logs failed for node {node}\n")
                    if output:
                        self.ui.error(output)

    def expire_crash(self):
        if self.config.crashexpireinterval == 0:
            return

        expirecrash = os.path.join(self.config.scriptsdir, "expire-crash")
        cmds = [(node, expirecrash, []) for node in self.config.hosts()]

        for node, success, output in self.executor.run_cmds(cmds):
            if not success:
                self.ui.error(f"expire-crash failed for node {node}\n")
                if output:
                    self.ui.error(output)

    def check_hosts(self):
        for host, status in self.executor.host_status():
            tag = f"alive-{host}"
            alive = status

            previous = self.config.get_state(tag)
            if previous is not None:
                if alive != previous:
                    self.pluginregistry.hostStatusChanged(host, alive)
                    if self.config.mailhostupdown:
                        up_or_down = "up" if alive else "down"
                        self.ui.info(f"host {host} {up_or_down}")

            self.config.set_state(tag, alive)

    def update_http_stats(self):
        if not self.config.statslogenable:
            return

        # Create meta file.
        if not os.path.exists(self.config.statsdir):
            try:
                os.makedirs(self.config.statsdir)
            except OSError as err:
                self.ui.error(
                    f"failure creating directory in zeekctl option statsdir: {err}"
                )
                return

            self.ui.info(f"creating directory for stats file: {self.config.statsdir}")

        metadat = os.path.join(self.config.statsdir, "meta.dat")
        try:
            with open(metadat, "w") as meta:
                for node in self.config.hosts():
                    meta.write(f"node {node} {node.type} {node.host}\n")

                meta.write(f"time {time.asctime()}\n")
                meta.write(f"version {self.config.version}\n")

                success, output = execute.run_localcmd("uname -a")
                if success and output:
                    # Note: "output" already has a '\n'
                    meta.write(f"os {output}")
                else:
                    meta.write("os <error>\n")

                success, output = execute.run_localcmd("hostname")
                if success and output:
                    # Note: "output" already has a '\n'
                    meta.write(f"host {output}")
                else:
                    meta.write("host <error>\n")

        except OSError as err:
            self.ui.error(f"failure creating file: {err}")
            return

        wwwdir = os.path.join(self.config.statsdir, "www")
        if not os.path.isdir(wwwdir):
            try:
                os.makedirs(wwwdir)
            except OSError as err:
                self.ui.error(f"failed to create directory: {err}")
                return

        # Update the WWW data
        statstocsv = os.path.join(self.config.scriptsdir, "stats-to-csv")

        success, output = execute.run_localcmd(
            f"{statstocsv} {self.config.statslog} {metadat} {wwwdir}"
        )
        if success:
            shutil.copy(metadat, wwwdir)
        else:
            self.ui.error(f"error reported by stats-to-csv\n{output}")

        # Append the current stats.log in spool to the one in ${statsdir}
        dst = os.path.join(self.config.statsdir, os.path.basename(self.config.statslog))
        try:
            with open(self.config.statslog) as fsrc:
                with open(dst, "a") as fdst:
                    shutil.copyfileobj(fsrc, fdst)
        except OSError as err:
            self.ui.error(f"failed to append file: {err}")
            return

        os.unlink(self.config.statslog)

    def run_cron_cmd(self):
        # Run external command if we have one.
        if self.config.croncmd:
            success, output = execute.run_localcmd(self.config.croncmd)
            if not success:
                self.ui.error(f"failure running croncmd: {self.config.croncmd}")



# This plugin sets necessary environment variables to run Zeek with
# PF_RING load balancing.

import ZeekControl.config
import ZeekControl.plugin


class LBPFRing(ZeekControl.plugin.Plugin):
    LB_POLICIES_ENV_MAP = {
        "2-tuple": "PCAP_PF_RING_USE_CLUSTER_PER_FLOW_2_TUPLE",
        "4-tuple": "PCAP_PF_RING_USE_CLUSTER_PER_FLOW_4_TUPLE",
        "5-tuple": "PCAP_PF_RING_USE_CLUSTER_PER_FLOW_5_TUPLE",
        "tcp-5-tuple": "PCAP_PF_RING_USE_CLUSTER_PER_FLOW_TCP_5_TUPLE",
        "6-tuple": "PCAP_PF_RING_USE_CLUSTER_PER_FLOW",
        "inner-2-tuple": "PCAP_PF_RING_USE_CLUSTER_PER_INNER_FLOW_2_TUPLE",
        "inner-4-tuple": "PCAP_PF_RING_USE_CLUSTER_PER_INNER_FLOW_4_TUPLE",
        "inner-5-tuple": "PCAP_PF_RING_USE_CLUSTER_PER_INNER_FLOW_5_TUPLE",
        "inner-tcp-5-tuple": "PCAP_PF_RING_USE_CLUSTER_PER_INNER_FLOW_TCP_5_TUPLE",
        "inner-6-tuple": "PCAP_PF_RING_USE_CLUSTER_PER_INNER_FLOW",
    }

    def __init__(self):
        super().__init__(apiversion=1)

    def name(self):
        return "lb_pf_ring"

    def pluginVersion(self):
        return 1

    def init(self):
        cluster_id = ZeekControl.config.Config.pfringclusterid
        if cluster_id == 0:
            return False

        pfringtype = ZeekControl.config.Config.pfringclustertype
        if pfringtype is None:
            pfringtype = "6-tuple"

        if pfringtype not in self.LB_POLICIES_ENV_MAP:
            self.error(f'PFRINGClusterType "{pfringtype}" not supported')
            return False

        pftype = self.LB_POLICIES_ENV_MAP[pfringtype]

        useplugin = False
        first_app_instance = ZeekControl.config.Config.pfringfirstappinstance
        app_instance = first_app_instance

        dd = {}
        for nn in self.nodes():
            if nn.type != "worker" or nn.lb_method != "pf_ring":
                continue

            useplugin = True

            if nn.host in dd:
                if nn.interface not in dd[nn.host]:
                    app_instance = first_app_instance
                    dd[nn.host][nn.interface] = cluster_id + len(dd[nn.host])
            else:
                app_instance = first_app_instance
                dd[nn.host] = {nn.interface: cluster_id}

            # Apply environment variables, but do not override values from
            # the node.cfg or zeekctl.cfg files.
            if pftype:
                nn.env_vars.setdefault(pftype, "1")

            if nn.interface.startswith("zc:"):
                # For the case where a user is doing RSS with ZC or
                # load-balancing with zbalance_ipc (through libpcap over
                # pf_ring)
                nn.env_vars.setdefault("PCAP_PF_RING_ZC_RSS", "1")
                nn.interface = f"{nn.interface}@{app_instance}"

            elif nn.interface.startswith("pf_ring::zc:"):
                # For the case where a user is doing RSS with ZC or
                # load-balancing with zbalance_ipc (through the zeek::pf_ring
                # plugin)
                nn.env_vars.setdefault("PCAP_PF_RING_ZC_RSS", "1")
                nn.interface = f"{nn.interface}@{app_instance}"

            elif nn.interface.startswith("dnacl"):
                # For the case where a user is running pfdnacluster_master (deprecated)
                nn.interface = f"{nn.interface}@{app_instance}"

            elif nn.interface.startswith("dna"):
                # For the case where a user is doing symmetric RSS with DNA (deprecated)
                nn.env_vars.setdefault("PCAP_PF_RING_DNA_RSS", "1")
                nn.interface = f"{nn.interface}@{app_instance}"

            else:
                nn.env_vars.setdefault(
                    "PCAP_PF_RING_CLUSTER_ID", dd[nn.host][nn.interface]
                )

            app_instance += 1
            nn.env_vars.setdefault("PCAP_PF_RING_APPNAME", f"zeek-{nn.interface}")

        return useplugin



import ast
import base64
import collections
import json
import logging
import os
import select
import subprocess
import time
import zlib
from queue import Empty, Queue
from threading import Thread


def get_muxer(shell):
    # The full path of the Python interpreter.  Configured by CMake.
    pythonpath = "/usr/local/bin/python3.14"

    muxer = r"""
import os,sys,subprocess,signal,select,json
TIMEOUT=120

def w(s):
	sys.stdout.write(repr(s) + "\n")
	sys.stdout.flush()

def exec_cmds(cmds):
	p=[]
	for i,cmd in enumerate(cmds):
		try:
			proc=subprocess.Popen(cmd,stdout=subprocess.PIPE,stderr=subprocess.PIPE __SHELL__)
			p.append((i,proc))
		except Exception as e:
			w((i,(1,'',str(e))))
	return p

w("ready")
commands=[]
signal.alarm(TIMEOUT)
for line in iter(sys.stdin.readline,"done\n"):
	commands.append(json.loads(line))

procs=exec_cmds(commands)
cmd_map={}
fd_map={}
fds=set()
for i,proc in procs:
	o={"idx":i, "proc":proc, "stdout":[], "stderr":[], "waiting":2}
	fd_map[proc.stdout]=o["stdout"]
	fd_map[proc.stderr]=o["stderr"]
	cmd_map[proc.stdout]=o
	cmd_map[proc.stderr]=o
	fds.update((proc.stdout,proc.stderr))

while fds:
	r,_,_=select.select(fds,[],[])
	for fd in r:
		output=os.read(fd.fileno(),1024)
		if output:
			fd_map[fd].append(output)
			continue

		cmd=cmd_map[fd]
		fds.remove(fd)
		cmd["waiting"]-=1
		if cmd["waiting"]:
			continue

		proc=cmd["proc"]
		status=proc.wait()
		out=b"".join(cmd["stdout"])
		err=b"".join(cmd["stderr"])
		w((cmd["idx"],(status,out,err)))

w("done")
"""

    if shell:
        muxer = muxer.replace("__SHELL__", ",shell=True")
    else:
        muxer = muxer.replace("__SHELL__", "")

    muxer = muxer.encode()
    muxer = base64.b64encode(zlib.compress(muxer))
    muxer = muxer.decode()
    muxer = f"{pythonpath} -c 'import zlib,base64; exec(zlib.decompress(base64.b64decode(b\"{muxer}\")))'\n"
    muxer = muxer.encode()

    return muxer


CmdResult = collections.namedtuple("CmdResult", "status stdout stderr")


class SSHMaster:
    def __init__(self, host, localaddrs):
        # The BatchMode=yes disables interactive prompting.  The LogLevel=error
        # prevents seeing login banners but allows error messages from ssh.
        self.base_cmd = [
            "ssh",
            "-o",
            "BatchMode=yes",
            "-o",
            "LogLevel=error",
            host,
        ]
        self.host = host
        self.need_connect = True
        self.master = None
        self.localaddrs = localaddrs
        self.run_mux = get_muxer(False)
        self.run_mux_shell = get_muxer(True)

    def connect(self):
        if self.need_connect:
            if self.host in self.localaddrs:
                cmd = ["sh"]
            else:
                cmd = self.base_cmd + ["sh"]
            self.master = subprocess.Popen(
                cmd,
                bufsize=0,
                stdout=subprocess.PIPE,
                stdin=subprocess.PIPE,
                close_fds=True,
                preexec_fn=os.setsid,
            )
            self.need_connect = False

    def readline_with_timeout(self, timeout):
        readable, _, _ = select.select([self.master.stdout], [], [], timeout)
        if not readable:
            return None
        jtxt = self.master.stdout.readline()
        jtxt = jtxt.decode()
        return jtxt

    def exec_command(self, cmd, shell=False, timeout=60):
        return self.exec_commands([cmd], shell, timeout)[0]

    def exec_commands(self, cmds, shell=False, timeout=60):
        self.send_commands(cmds, timeout, shell)
        return self.collect_results(timeout)

    def send_commands(self, cmds, timeout, shell=False):
        self.connect()
        if shell:
            self.master.stdin.write(self.run_mux_shell)
        else:
            self.master.stdin.write(self.run_mux)
        self.master.stdin.flush()

        # Wait until we receive the "ready" message from muxer script
        self.readline_with_timeout(timeout)

        for cmd in cmds:
            jcmd = f"{json.dumps(cmd)}\n"
            jcmd = jcmd.encode()
            self.master.stdin.write(jcmd)
        # Note: the "b" string prefix here for Py3 is ignored by Py2.6-2.7
        self.master.stdin.write(b"done\n")
        self.master.stdin.flush()
        self.sent_commands = len(cmds)

    def collect_results(self, timeout):
        outputs = [
            Exception(f"Command timeout on host {self.host}")
        ] * self.sent_commands

        while True:
            line = self.readline_with_timeout(timeout)
            if not line:
                logging.debug("Command timeout on host %s", self.host)
                self.close()
                break
            resp = ast.literal_eval(line)
            if resp == "done":
                break
            idx, result = resp
            status, out, err = result

            out = out.decode(errors="replace")
            err = err.decode(errors="replace")

            outputs[idx] = CmdResult(status, out, err)
        return outputs

    def close(self):
        if not self.master:
            return
        self.master.stdin.close()
        try:
            self.master.kill()
        except OSError:
            pass
        self.master.wait()
        self.master = None
        self.need_connect = True

    __del__ = close


STOP_RUNNING = object()


class HostHandler(Thread):
    def __init__(self, host, localaddrs, timeout):
        self.host = host
        self.localaddrs = localaddrs
        self.timeout = timeout
        self.q = Queue()
        self.alive = False
        self.master = None
        Thread.__init__(self)

    def shutdown(self):
        self.q.put((STOP_RUNNING, None, None))

    def connect(self):
        if self.master:
            self.master.close()
        self.master = SSHMaster(self.host, self.localaddrs)

    def ping(self):
        # Error message should indicate whether or not ssh is being used.
        msgstr = "" if self.host in self.localaddrs else "ssh "

        # Error message shows if a connection was previously established.
        if self.alive:
            msg = f"Lost {msgstr}connection to host {self.host}"
        else:
            msg = f"Failed to establish {msgstr}connection to host {self.host}"

        # This will be set to True below only if the "ping" is received.
        self.alive = False

        try:
            resp = self.master.exec_command(["/bin/echo", "ping"], timeout=10)
        except Exception as e:
            # This happens most likely due to broken pipe (i.e., ssh
            # terminates, usually because it couldn't connect, or its own
            # timeout occurred).
            return f"{msg}: {e}"

        try:
            ping_recvd = resp.stdout.strip() == "ping"
        except Exception:
            # This happens when there was a timeout in SSHMaster (in this
            # situation, that almost always means unable to establish
            # connection or a loss of connection).
            return msg

        if ping_recvd:
            self.alive = True
            return ""

        # This should probably never happen.
        return f"Communication failure with host {self.host} when checking connection"

    def connect_and_ping(self):
        if not self.alive:
            self.connect()
        return self.ping()

    def run(self):
        while True:
            if self.iteration():
                return

    def iteration(self):
        try:
            item, shell, rq = self.q.get(timeout=30)
        except Empty:
            self.connect_and_ping()
            return False

        if item is STOP_RUNNING:
            return True

        msg = self.connect_and_ping()
        if not self.alive:
            logging.debug(msg)
            resp = [Exception(msg)] * len(item)
            rq.put(resp)
            return False

        try:
            resp = self.master.exec_commands(item, shell, self.timeout)
        except Exception as e:
            self.alive = False
            msgstr = "" if self.host in self.localaddrs else "ssh "
            msg = f"Lost {msgstr}connection while running command on host {self.host}: {e}"
            logging.debug(msg)
            resp = [Exception(msg)] * len(item)
            time.sleep(2)
        rq.put(resp)

        return False

    def send_commands(self, commands, shell, rq):
        self.q.put((commands, shell, rq))


class MultiMasterManager:
    def __init__(self, localaddrs=[]):
        self.masters = {}
        self.response_queues = {}
        self.localaddrs = localaddrs

    def setup(self, host, timeout):
        if host not in self.masters:
            self.masters[host] = HostHandler(host, self.localaddrs, timeout)
            self.masters[host].start()

    def send_commands(self, host, commands, timeout, shell=False):
        self.setup(host, timeout)
        rq = Queue()
        self.response_queues[host] = rq
        self.masters[host].send_commands(commands, shell, rq)

    def get_result(self, host, hosttimeout):
        # Add a few seconds to the host timeout in order to let the
        # command timeout happen first.
        hosttimeout += 5

        rq = self.response_queues[host]
        try:
            return rq.get(timeout=hosttimeout)
        except Empty:
            self.shutdown(host)
            # This can happen due to commands that take a while to run, a
            # loss of connectivity to remote host, or both.
            return [
                Exception(f"Timeout waiting for commands to finish on host {host}")
            ]  # FIXME: needs to be the right length

    def exec_command(self, host, command, timeout=30):
        return self.exec_commands(host, [command], timeout)[0]

    def exec_commands(self, host, commands, timeout=60):
        self.send_commands(host, commands, timeout)
        return self.get_result(host, timeout)

    def exec_multihost_commands(self, cmds, shell=False, timeout=60):
        hosts = collections.defaultdict(list)
        for host, cmd in cmds:
            hosts[host].append(cmd)

        for host, cmds in hosts.items():
            self.send_commands(host, cmds, timeout, shell)

        for host in hosts:
            for res in self.get_result(host, timeout):
                yield host, res

    def host_status(self):
        for h, o in self.masters.items():
            if h not in self.localaddrs:
                yield h, o.alive

    def shutdown(self, host):
        self.masters[host].shutdown()
        del self.masters[host]

    def shutdown_all(self):
        for handler in self.masters.values():
            handler.shutdown()
        self.masters = {}

    __del__ = shutdown_all



import errno
import os


# 'src' is the file to which the link will point, and 'dst' is the link to make
def force_symlink(src, dst):
    try:
        os.symlink(src, dst)
    except OSError as e:
        if e.errno == errno.EEXIST:
            os.remove(dst)
            os.symlink(src, dst)
        else:
            raise


# Returns an IP address string suitable for embedding in a Zeek script,
# for IPv6 colon-hexadecimal address strings, that means surrounding it
# with square brackets.
def format_zeek_addr(addr):
    if ":" not in addr:
        return addr
    else:
        return f"[{addr}]"


# Returns an IP prefix string suitable for embedding in a Zeek script,
# for IPv6 colon-hexadecimal prefix strings, that means surrounding the
# IP address part with square brackets.
def format_zeek_prefix(prefix):
    if ":" not in prefix:
        return prefix
    else:
        parts = prefix.split("/")
        return f"[{parts[0]}]/{parts[1]}"


# Returns an IP address string suitable for use with rsync, which requires
# encasing IPv6 addresses in square brackets, and some shells may require
# quoting the brackets.
def format_rsync_addr(addr):
    if ":" not in addr:
        return addr
    else:
        return f"'[{addr}]'"


# Convert a number into a string with a unit (e.g., 1024 into "1K").
def number_unit_str(num):
    units = (("G", 1024 * 1024 * 1024), ("M", 1024 * 1024), ("K", 1024))
    for unit, factor in units:
        if num >= factor:
            return f"{num / factor:3.0f}{unit}"
    return f" {num:3.0f}"



#
# Generates the dynamic parts of the ZeekControl documentation.
#

import inspect


def api(*deco_args):
    if len(deco_args) == 1 and callable(deco_args[0]):
        # No argument to decorator.
        method = deco_args[0]
        method._doc = ""
        return method

    else:
        # Tag argument.
        def _api(method):
            method._doc = deco_args[0]
            return method

        return _api


def print_indented(text, level):
    out = ""
    if not isinstance(text, list):
        text = text.splitlines()

    for line in text:
        if line:
            out += "{} {}\n".format("    " * level, line)
        else:
            out += "\n"

    out += "\n"

    return out


# Prints API documentation for a class. Includes all methods tagged with
# @api(tag). (Use an unknown tag to not exclude all methods.) If header is
# False, the class's name and doc string is not included.
def print_class(cls, tag="", header=True):
    out = ""
    methods = {}

    for key, val in cls.__dict__.items():
        if not inspect.isfunction(val):
            continue

        if not "_doc" in val.__dict__:
            continue

        if val.__dict__["_doc"] == tag:
            methods[key] = val

    if header:
        out += f".. _{cls.__name__}:\n\n"
        out += f"Class ``{cls.__name__}``\n"
        out += "~~~~~~~~%s~~" % ("~" * len(cls.__name__))
        out += "\n\n"
        out += f"class **{cls.__name__}**\n"
        out += print_indented(inspect.getdoc(cls), 1)

    for name in sorted(methods.keys()):
        func = methods[name]

        spec = inspect.getfullargspec(func)

        out += print_indented(f".. _{cls.__name__}.{name}:", 1)
        out += print_indented("**{}** ({})".format(name, ", ".join(spec.args)), 1)
        out += print_indented(inspect.getdoc(func), 2)

    return out



#
# One ZeekControl node.
#

import copy
import os

from ZeekControl import doc


class Node:
    """Class representing one node of the ZeekControl maintained setup. In
    standalone mode, there's always exactly one node of type ``standalone``. In
    a cluster setup, there is zero or more of type ``logger``, exactly one of
    type ``manager``, one or more of type ``proxy``, and zero or more of type
    ``worker``.  The manager will handle writing logs if there are no loggers
    defined in a cluster.

    A ``Node`` object has a number of keys with values that are set via the
    ``node.cfg`` file and can be accessed directly (from a plugin) via
    corresponding Python attributes (e.g., ``node.name``):

        ``name`` (string)
            The name of the node, which corresponds to the ``[<name>]``
            section in ``node.cfg``.

        ``type`` (string)
            The type of the node.  In a standalone configuration, the only
            allowed type is ``standalone``.  In a cluster configuration, the
            type must be one of: ``logger``, ``manager``, ``proxy``,
            or ``worker``.

        ``host`` (string)
            The hostname or IP address of the system the node is
            running on.  Every node must specify a host.

        ``interface`` (string)
            The network interface for the Zeek worker (or standalone node) to
            use; empty if not set.

        ``lb_procs`` (integer)
            The number of clustered Zeek workers you'd like to start up.  If
            specified, this number must be greater than zero and a load
            balancing method must also be specified.  This option is valid only
            for worker nodes.

        ``lb_method`` (string)
            The load balancing method to distribute packets to all of the
            Zeek workers.  This must be one of: ``af_packet``, ``pf_ring``,
            ``myricom``, ``custom``, or ``interfaces``.  This option can have
            a value only if the ``lb_procs`` option has a value.

        ``lb_interfaces`` (string)
            A comma-separated list of network interface names for the Zeek
            worker to use.  The number of interfaces in this list must
            equal the value of the ``lb_procs`` option.

            This option can be specified only when the load balancing method
            is ``interfaces``.

        ``pin_cpus`` (string)
            A comma-separated list of CPU numbers to which the node's Zeek
            processes will be pinned.  If not specified, then CPU pinning will
            not be used for this node.  This option is supported only on
            Linux and FreeBSD, and is ignored on all other platforms.

            CPU numbering starts at zero (e.g.,
            the only valid CPU numbers for a machine with one dual-core
            processor would be 0 and 1).  If the length of this list does not
            match the number of Zeek processes for this node, then some CPUs
            could have zero (if too many CPU numbers are specified) or more
            than one (if not enough CPU numbers are specified) Zeek processes
            pinned to them.  Only the specified CPU numbers will be used,
            regardless of whether additional CPU cores exist.

        ``env_vars`` (string)
            A comma-separated list of environment variables to set when
            running Zeek (e.g., ``env_vars=VAR1=1,VAR2=2``).  These
            node-specific values override any global values specified in
            the ``zeekctl.cfg`` file.

        ``aux_scripts`` (string)
            Any node-specific Zeek script configured for this node.

        ``zone_id`` (string)
            If ZeekControl is managing a cluster comprised of nodes
            using non-global IPv6 addresses, then this configures the
            :rfc:`4007` ``zone_id`` string that the node associates with
            the common zone that all cluster nodes are a part of.  This
            identifier may differ between nodes.

    Any attribute that is not defined in ``node.cfg`` will be empty.

    In addition, plugins can override `Plugin.nodeKeys`_ to define their own
    node keys, which can then be likewise set in ``node.cfg``. The key names
    will be prepended with the plugin's `Plugin.prefix`_ (e.g., for the plugin
    ``test``, the node key ``foo`` is set by adding ``test.foo=value`` to
    ``node.cfg``).

    Finally, a Node object has the following methods that can be called
    from a plugin:
    """

    # Valid keys in nodes file. The values will be stored in attributes of the
    # same name. Custom keys can be add via addKey().
    _keys = {
        "type": 1,
        "host": 1,
        "interface": 1,
        "aux_scripts": 1,
        "zeekbase": 1,
        "ether": 1,
        "zone_id": 1,
        "lb_procs": 1,
        "lb_method": 1,
        "lb_interfaces": 1,
        "pin_cpus": 1,
        "env_vars": 1,
        "count": 1,
    }

    def __init__(self, config, name):
        """Instantiates a new node of the given name."""
        self.name = name
        self._config = config

        for key in Node._keys:
            self.__dict__[key] = ""

    def __str__(self):
        return self.name

    def __format__(self, format_str):
        return format(self.name, format_str)

    def copy(self):
        n = Node(self._config, self.name)

        for key in self.__dict__:
            if key.startswith("_"):
                # This is to avoid copying _config, which causes problems.
                setattr(n, key, getattr(self, key))
            else:
                # Must make a copy of some config items (e.g. env_vars) so that
                # changes to the value only affect one node.
                setattr(n, key, copy.copy(getattr(self, key)))

        return n

    def items(self):
        """Returns a list of (key, value) tuples, sorted by key, of a node."""

        def tostr(v):
            if isinstance(v, dict):
                return ",".join([f"{key}={val}" for (key, val) in sorted(v.items())])
            else:
                return str(v)

        return [(k, tostr(self.__dict__[k])) for k in sorted(self.__dict__.keys())]

    @doc.api
    def describe(self):
        """Returns an extended string representation of the node including all
        its keys with values (sorted by key)."""

        def fmt(v):
            if isinstance(v, list):
                v = ",".join(v)
            elif isinstance(v, dict):
                v = ",".join([f"{key}={val}" for (key, val) in sorted(v.items())])

            return v

        # Do not output attributes starting with underscore, because they are
        # for internal use and don't provide useful information to the user.
        return f"{self.name:>16s} - " + " ".join(
            [
                f"{k}={fmt(self.__dict__[k])}"
                for k in sorted(self.__dict__.keys())
                if not k.startswith("_")
            ]
        )

    def to_dict(self):
        d = dict(self.items())
        d["name"] = self.name
        d["description"] = self.describe()
        return d

    @doc.api
    def cwd(self):
        """Returns a string with the node's working directory."""
        return os.path.join(self._config.spooldir, self.name)

    def setPID(self, pid):
        """Stores the process ID of the node's Zeek process."""
        key = f"{self.name}-pid"
        self._config.set_state(key, pid)
        key = f"{self.name}-host"
        self._config.set_state(key, self.host)

    @doc.api
    def getPID(self):
        """Returns the process ID of the node's Zeek process if running, and
        None otherwise."""
        key = f"{self.name}-pid"
        return self._config.get_state(key)

    def clearPID(self):
        """Clears the stored process ID for the node's Zeek process, indicating
        that it is no longer running."""
        key = f"{self.name}-pid"
        self._config.set_state(key, None)

    def setCrashed(self):
        """Marks node's Zeek process as having terminated unexpectedly."""
        key = f"{self.name}-crashed"
        self._config.set_state(key, True)

    def clearCrashed(self):
        """Clears the mark for the node's Zeek process having terminated
        unexpectedly."""
        key = f"{self.name}-crashed"
        self._config.set_state(key, False)

    @doc.api
    def hasCrashed(self):
        """Returns True if the node's Zeek process has exited abnormally."""
        key = f"{self.name}-crashed"
        val = self._config.get_state(key)
        if val is None:
            val = False
        return val

    def getExpectRunning(self):
        """Returns True if we expect the node's Zeek process to be running."""
        key = f"{self.name}-expect-running"
        val = self._config.get_state(key)
        if val is None:
            val = False
        return val

    def setExpectRunning(self, val):
        key = f"{self.name}-expect-running"
        self._config.set_state(key, val)

    def setPort(self, port):
        """Set the Zeek port this node is using."""
        key = f"{self.name}-port"
        self._config.set_state(key, port)

    @doc.api
    def getPort(self):
        """Returns an integer with the port number that this node's
        communication system is listening on for incoming connections, or -1 if
        no such port has been set yet.
        """
        key = f"{self.name}-port"
        return self._config.get_state(key) or -1

    @staticmethod
    def addKey(kw):
        """Adds a supported node key. This is used by the PluginRegistry to
        register custom keys."""

        # We need to convert to lowercase here because Python's configparser
        # automatically converts keys to lowercase when reading node.cfg.
        Node._keys[kw.lower()] = 1


# The sorting order for node types used by the sorting functions below
# (specifying the sort order explicitly is useful if we don't want to
# use alphabetical order).
_typeorder = ("standalone", "logger", "manager", "proxy", "worker")


# Sorting key function for a list of nodes.
def sortnode(n):
    try:
        return _typeorder.index(n.type), n.count
    except ValueError:
        return len(_typeorder), n.count


# Sorting key function for a list of tuples, where the first tuple element is
# a node.
def sorttuple(t):
    n = t[0]
    try:
        return _typeorder.index(n.type), n.count
    except ValueError:
        return len(_typeorder), n.count


# Given a list of nodes (all of the same type), return a string that describes
# the list of nodes (in either singular or plural form).  This string is
# just for informational output, and doesn't have any other use or meaning.
# For standalone node type, the node name is returned instead.
def nodes_describe(nodes):
    nodetype = nodes[0].type

    if nodetype == "standalone":
        return nodes[0].name
    elif nodetype == "manager":
        return "manager"
    elif nodetype == "logger":
        return "logger%s" % ("" if len(nodes) == 1 else "s")
    elif nodetype == "proxy":
        return "prox%s" % ("y" if len(nodes) == 1 else "ies")
    elif nodetype == "worker":
        return "worker%s" % ("" if len(nodes) == 1 else "s")


# Return a list of all node types.
def node_types():
    return ["logger", "manager", "proxy", "worker", "standalone"]


# Check if the given node is a certain type.
def is_standalone(n):
    return n.type == "standalone"


def is_manager(n):
    return n.type == "manager"


def is_logger(n):
    return n.type == "logger"


def is_proxy(n):
    return n.type == "proxy"


def is_worker(n):
    return n.type == "worker"


# Given a list of nodes, return separate lists for each type of node.
def separate_types(nodes):
    loggers = []
    manager = []
    proxies = []
    workers = []

    for n in nodes:
        if n.type == "worker":
            workers += [n]
        elif n.type == "proxy":
            proxies += [n]
        elif n.type in ("manager", "standalone"):
            manager += [n]
        elif n.type == "logger":
            loggers += [n]

    return loggers, manager, proxies, workers


# Map of node groups to node types (here, "_ALL_" is for internal use only and
# matches all node types).
grouptype = {
    "all": "_ALL_",
    "loggers": "logger",
    "manager": "manager",
    "proxies": "proxy",
    "workers": "worker",
}


# Return a list of all node groups.  These are for convenience when using
# zeekctl commands (e.g. "zeekctl start workers").
def node_groups():
    return list(grouptype.keys())


# Return the node type (or "_ALL_", which matches all node types) of a
# specified group name.  If the "tag" doesn't match any group name, then None
# is returned.
def group_type(tag):
    return grouptype.get(tag)


# Return the name of a node group.
def manager_group():
    return "manager"


def logger_group():
    return "loggers"


def proxy_group():
    return "proxies"


def worker_group():
    return "workers"



# Functions to install files on all nodes.

import binascii
import os

from ZeekControl import config, util

# In all paths given in this file, ${<option>} will replaced with the value of
# the corresponding configuration option.


# Directories/files in form (path, mirror, optional) which are synced from the manager to
# all remote hosts.
# If "mirror" is False, then the path is assumed to be a directory and it will
# just be created on the remote host.  If "mirror" is True, then the path
# is fully mirrored recursively.  If "optional" is True, then it's ok for
# the path to not exist on the manager.
def get_syncs():
    syncs = [
        ("${zeekbase}", False, False),
        ("${zeekbase}/share", True, False),
        ("${cfgdir}", True, False),
        ("${libdir}", True, True),
        ("${libdir64}", True, True),
        ("${bindir}", True, False),
        ("${policydirsiteinstall}", True, False),
        ("${policydirsiteinstallauto}", True, False),
        # ("${policydir}", True, False),
        # ("${staticdir}", True, False),
        ("${logdir}", False, False),
        ("${spooldir}", False, False),
        ("${fileextractdir}", False, False),
        ("${tmpdir}", False, False),
        ("${zeekctlconfigdir}/zeekctl-config.sh", True, False),
    ]

    return syncs


# In NFS-mode, only these will be synced.
def get_nfssyncs():
    nfssyncs = [
        ("${spooldir}", False, False),
        ("${fileextractdir}", False, False),
        ("${tmpdir}", False, False),
        ("${policydirsiteinstall}", True, False),
        ("${policydirsiteinstallauto}", True, False),
        ("${zeekctlconfigdir}/zeekctl-config.sh", True, False),
    ]

    return nfssyncs


# Determine relative cfg_path
def splitall(path):
    """Break out all parts of a path into a list"""
    allparts = []
    while True:
        parts = os.path.split(path)
        if parts[0] == path:
            allparts.insert(0, parts[0])
            break
        if parts[1] == path:
            allparts.insert(0, parts[1])
            break
        path = parts[0]
        allparts.insert(0, parts[1])
    return allparts


def relpath(src, dst):
    """Calculate the relative path to dst)"""
    srcparts = splitall(src)
    dstparts = splitall(dst)
    while srcparts and dstparts and srcparts[0] == dstparts[0]:
        srcparts.pop(0)
        dstparts.pop(0)
    relparts = (len(dstparts) - 1) * [".."] + srcparts
    return os.path.join(*relparts)


# Generate a shell script "zeekctl-config.sh" that sets env. vars. that
# correspond to zeekctl config options.
def make_zeekctl_config_sh(cmdout):
    ostr = ""
    for varname, value in config.Config.options(dynamic=False):
        if isinstance(value, bool):
            # Convert bools to the string "1" or "0"
            value = "1" if value else "0"
        else:
            value = str(value)

        # In order to prevent shell errors, here we convert plugin
        # option names to use underscores, and double quotes in the value
        # are escaped.
        ostr += '{}="{}"\n'.format(varname.replace(".", "_"), value.replace('"', '\\"'))

    # Rather than just overwriting the file, we first write out a tmp file,
    # and then rename it to avoid a race condition where a process outside of
    # zeekctl (such as archive-log) is trying to read the file while it is
    # being written.
    cfg_path = os.path.join(config.Config.zeekctlconfigdir, "zeekctl-config.sh")
    tmp_path = os.path.join(config.Config.zeekctlconfigdir, ".zeekctl-config.sh.tmp")

    try:
        with open(tmp_path, "w") as out:
            out.write(ostr)
    except OSError as e:
        cmdout.error(f"failed to write file: {e}")
        return False

    try:
        os.rename(tmp_path, cfg_path)
    except OSError as e:
        cmdout.error(f"failed to rename file {tmp_path}: {e}")
        return False

    symlink = os.path.join(config.Config.scriptsdir, "zeekctl-config.sh")

    # Use a relative path instead of absolute (at request/usage of the FreeBSD
    # port, see https://github.com/zeek/zeekctl/pull/22)
    sym_cfg_path = relpath(cfg_path, symlink)

    # check if the symlink needs to be updated
    try:
        update_link = (
            not os.path.islink(symlink) or os.readlink(symlink) != sym_cfg_path
        )
    except OSError as e:
        cmdout.error(f"failed to read symlink: {e}")
        return False

    if update_link:
        # attempt to update the symlink
        try:
            util.force_symlink(sym_cfg_path, symlink)
        except OSError as e:
            cmdout.error(
                f"failed to update symlink '{symlink}' to point to '{sym_cfg_path}': {e.strerror}"
            )
            return False

    return True


# Create Zeek-side zeekctl configuration file.
def make_layout(path, cmdout, silent=False):
    class Port:
        def __init__(self, startport):
            # This is the first port number to use.
            self.p = startport

        # Record the port number that the specified node will use (if node is
        # None, then don't record it) and return that port number.
        def use_port(self, node):
            port = self.p
            # Increment the port number, since we're using the current one.
            self.p += 1

            if node is not None:
                node.setPort(port)

            return port

    manager = config.Config.manager()
    zeekport = Port(config.Config.zeekport)
    metricsport = None
    if config.Config.metricsport != 0:
        metricsport = Port(config.Config.metricsport)

    if config.Config.standalone:
        if not silent:
            cmdout.info("generating standalone-layout.zeek ...")

        filename = os.path.join(path, "standalone-layout.zeek")

        ostr = "# Automatically generated. Do not edit.\n"
        # This is the port that standalone nodes listen on for remote
        # control by default.
        ostr += f"redef Broker::default_port = {zeekport.use_port(manager)}/tcp;\n"
        if metricsport:
            ostr += '@if ( getenv("ZEEKCTL_DISABLE_LISTEN") == "" )\n'
            ostr += (
                f"redef Telemetry::metrics_port = {metricsport.use_port(None)}/tcp;\n"
            )
            ostr += "@endif\n"
        ostr += "\n"
        ostr += "event zeek_init()\n"
        ostr += "\t{\n"
        ostr += '\tif ( getenv("ZEEKCTL_DISABLE_LISTEN") == "" )\n'
        ostr += "\t\tBroker::listen();\n"
        ostr += "\t}\n"

    else:
        if not silent:
            cmdout.info("generating cluster-layout.zeek ...")

        filename = os.path.join(path, "cluster-layout.zeek")
        workers = config.Config.workers()
        proxies = config.Config.proxies()
        loggers = config.Config.loggers()

        # If no loggers are defined, then manager does the logging.
        manager_is_logger = "F" if loggers else "T"

        ostr = "# Automatically generated. Do not edit.\n"
        ostr += f"redef Cluster::manager_is_logger = {manager_is_logger};\n"
        ostr += "redef Cluster::nodes = {\n"

        # Control definition.  For now just reuse the manager information.
        ostr += f'\t["control"] = [$node_type=Cluster::CONTROL, $ip={util.format_zeek_addr(manager.addr)}, $p={zeekport.use_port(None)}/tcp],\n'

        # Manager definition. Manager should come first so that the metrics port
        # set in zeekctl.cfg will be the one used for the manager.
        ostr += f'\t["{manager.name}"] = [$node_type=Cluster::MANAGER, $ip={util.format_zeek_addr(manager.addr)}, $p={zeekport.use_port(manager)}/tcp'
        if metricsport:
            ostr += f", $metrics_port={metricsport.use_port(None)}/tcp"
        ostr += "],\n"

        # Loggers definition
        for lognode in loggers:
            ostr += f'\t["{lognode.name}"] = [$node_type=Cluster::LOGGER, $ip={util.format_zeek_addr(lognode.addr)}, $p={zeekport.use_port(lognode)}/tcp'
            if metricsport:
                ostr += f", $metrics_port={metricsport.use_port(None)}/tcp"
            ostr += "],\n"

        # Proxies definition (all proxies use same logger as the manager)
        for p in proxies:
            ostr += f'\t["{p.name}"] = [$node_type=Cluster::PROXY, $ip={util.format_zeek_addr(p.addr)}, $p={zeekport.use_port(p)}/tcp, $manager="{manager.name}"'
            if metricsport:
                ostr += f", $metrics_port={metricsport.use_port(None)}/tcp"
            ostr += "],\n"

        # Workers definition
        for w in workers:
            p = w.count % len(proxies)
            ostr += f'\t["{w.name}"] = [$node_type=Cluster::WORKER, $ip={util.format_zeek_addr(w.addr)}, $p={zeekport.use_port(w)}/tcp, $manager="{manager.name}"'
            if metricsport:
                ostr += f", $metrics_port={metricsport.use_port(None)}/tcp"
            ostr += "],\n"

        # Activate time-machine support if configured.
        if config.Config.timemachinehost:
            ostr += f'\t["time-machine"] = [$node_type=Cluster::TIME_MACHINE, $ip={config.Config.timemachinehost}, $p={config.Config.timemachineport}],\n'

        ostr += "};\n"

    # If UseWebSocket is True, render a zeek_init() handler that starts a WebSocket
    # server on the manager node.
    if config.Config.usewebsocket:
        ostr += "\n"
        ostr += "event zeek_init()\n\t{\n"
        ostr += '\tif ( getenv("ZEEKCTL_DISABLE_LISTEN") == "" && (Cluster::local_node_type() == Cluster::MANAGER || ! Cluster::is_enabled()) )\n'
        ostr += "\t\t{\n"
        ostr += f"\t\tCluster::listen_websocket([$listen_addr={config.Config.websockethost}, $listen_port={config.Config.websocketport}/tcp]);\n"
        ostr += "\t\t}\n"
        ostr += "\t}\n"

    try:
        with open(filename, "w") as out:
            out.write(ostr)
    except OSError as e:
        cmdout.error(f"failed to write file: {e}")
        return False

    return True


# Reads in a list of networks from file.
def read_networks(fname):
    nets = []

    with open(fname) as f:
        for line in f:
            line = line.strip()
            if not line or line.startswith("#"):
                continue

            fields = line.split(None, 1)

            cidr = util.format_zeek_prefix(fields[0])
            tag = fields[1] if len(fields) == 2 else ""

            nets += [(cidr, tag)]

    return nets


# Create Zeek script which contains a list of local networks.
def make_local_networks(path, cmdout):
    netcfg = config.Config.localnetscfg

    try:
        nets = read_networks(netcfg)
    except IndexError:
        cmdout.error(f"invalid CIDR notation in file: {netcfg}")
        return False
    except OSError as e:
        cmdout.error(f"failed to read file: {e}")
        return False

    ostr = "# Automatically generated. Do not edit.\n\n"

    ostr += "redef Site::local_nets = {\n"
    for cidr, tag in nets:
        ostr += f"\t{cidr},"
        if tag:
            ostr += f"\t# {tag}"
        ostr += "\n"
    ostr += "};\n\n"

    try:
        with open(os.path.join(path, "local-networks.zeek"), "w") as out:
            out.write(ostr)
    except OSError as e:
        cmdout.error(f"failed to write file: {e}")
        return False

    return True


def make_zeekctl_config_policy(path, cmdout, plugin_reg):
    ostr = "# Automatically generated. Do not edit.\n"
    ostr += f'redef Notice::mail_dest = "{config.Config.mailto}";\n'
    ostr += (
        f'redef Notice::mail_dest_pretty_printed = "{config.Config.mailalarmsto}";\n'
    )
    ostr += f'redef Notice::sendmail = "{config.Config.sendmail}";\n'
    ostr += (
        f'redef Notice::mail_subject_prefix = "{config.Config.mailsubjectprefix}";\n'
    )
    ostr += f'redef Notice::mail_from = "{config.Config.mailfrom}";\n'
    ostr += f'redef Broker::table_store_db_directory = "{config.Config.brokerdbdir}";\n'
    if not config.Config.standalone:
        loggers = config.Config.loggers()
        ntype = "LOGGER" if loggers else "MANAGER"
        ostr += f"@if ( Cluster::local_node_type() == Cluster::{ntype} )\n"

    ostr += f"redef Log::default_rotation_interval = {config.Config.logrotationinterval} secs;\n"
    ostr += f"redef Log::default_mail_alarms_interval = {config.Config.mailalarmsinterval} secs;\n"

    if not config.Config.standalone:
        ostr += "@endif\n"

    ostr += f"redef Pcap::snaplen = {config.Config.pcapsnaplen};\n"
    ostr += f"redef Pcap::bufsize = {config.Config.pcapbufsize};\n"

    if not config.Config.privateaddressspaceislocal:
        ostr += "redef Site::private_address_space_is_local = F;\n"

    seed_str = make_global_hash_seed()
    ostr += f'redef global_hash_seed = "{seed_str}";\n'

    ostr += f'redef Cluster::default_store_dir = "{config.Config.defaultstoredir}";\n'

    if config.Config.compresslogsinflight > 0:
        ostr += f"redef LogAscii::gzip_level = {config.Config.compresslogsinflight};\n"
        ostr += f'redef LogAscii::gzip_file_extension = "{config.Config.compressextension}";\n'

    if config.Config.fileextractdir:
        ostr += f'redef FileExtract::prefix = "{config.Config.fileextractdir}/" + Cluster::node;\n'

    if config.Config.metricsaddress:
        # Telemetry::metrics_address is a string, so need extra quotes.
        #
        # Not that if Telemetry::metrics_address is already set, e.g via the
        # ZEEK_DEFAULT_LISTEN_ADDRESS environment variable or an explicit
        # redef in local.zeek, then MetricsAddress has no effect.
        ostr += "@if ( |Telemetry::metrics_address| == 0 )\n"
        ostr += (
            f'redef Telemetry::metrics_address = "{config.Config.metricsaddress}";\n'
        )
        ostr += "@endif\n"

    ostr += plugin_reg.getZeekctlConfig(cmdout)

    filename = os.path.join(path, "zeekctl-config.zeek")
    try:
        with open(filename, "w") as out:
            out.write(ostr)
    except OSError as e:
        cmdout.error(f"failed to write file: {e}")
        return False

    return True


# Create a new random seed value if one is not found in the state database (this
# ensures a consistent value after a restart).  Return a string representation.
def make_global_hash_seed():
    seed_str = config.Config.get_state("global-hash-seed")

    if not seed_str:
        # Get 4 bytes of random data (Zeek uses 4 bytes to create an initial
        # seed in the Hasher::MakeSeed() function if the Zeek script constant
        # global_hash_seed is an empty string).
        seed = os.urandom(4)

        # Convert each byte of seed value to a two-digit hex string.
        seed_str = binascii.hexlify(seed)
        seed_str = seed_str.decode()

        config.Config.set_state("global-hash-seed", seed_str)

    return seed_str



import json
import os.path
import subprocess
from pathlib import Path

import ZeekControl.config
import ZeekControl.plugin
from ZeekControl.exceptions import ConfigurationError


class ClusterBackendZeroMQ(ZeekControl.plugin.Plugin):
    def __init__(self):
        super().__init__(apiversion=1)

    def name(self):
        return "cluster_backend_zeromq"

    def pluginVersion(self):
        return 1

    def options(self):
        return [
            (
                "disable_unencrypted_warning",
                "bool",
                False,
                "Disable the multi-node unencrypted warning.",
            ),
            (
                "use_curve_encryption",
                "string",
                "auto",
                "Whether to enable ZeroMQ CURVE-based encryption for cluster communication (auto, 0 or 1)",
            ),
            (
                "curve_dir",
                "string",
                "${SpoolDir}/zeromq/curve",
                "Directory to persistently store client and server curve keys on the manager",
            ),
        ]

    def generate_keypair(self, publickey: Path, secretkey: Path):
        """
        Generate the ZeroMQ CURVE keypair in Z85 encoded format and write
        them to the files pointed at by publickey and secretkey. The keys
        themselves are generated using zeek and calling the appropriate BiF.
        """
        zeek = self.getGlobalOption("zeek")
        if not os.path.lexists(zeek):
            raise ConfigurationError(f"cannot find Zeek binary: {zeek}")

        args = [
            zeek,
            "-b",
            "-e",
            "print to_json(Cluster::Backend::ZeroMQ::generate_keypair())",
        ]
        output = subprocess.check_output(args)
        loaded = json.loads(output)
        public, secret = loaded["public"], loaded["secret"]
        if len(public) != 40 or len(secret) != 40:
            raise ConfigurationError(
                f"failed to create ZeroMQ CURVE keypair {loaded!r}"
            )

        publickey.touch(exist_ok=True)
        publickey.chmod(0o600)
        publickey.write_text(public + "\n")

        secretkey.touch(exist_ok=True)
        secretkey.chmod(0o600)
        secretkey.write_text(secret + "\n")

    def init(self):
        """
        Enable the plugin if ClusterBackend setting is ZeroMQ.
        """
        backend = self.getGlobalOption("ClusterBackend")
        if backend.lower() != "zeromq":
            return False

        # Switch topic separator for node topics to "." as that's what
        # the ZeroMQ cluster backend uses.
        if not self.getGlobalOption("ClusterTopicSeparator"):
            ZeekControl.config.Config.set_option("ClusterTopicSeparator", ".")

        # Lookup ZeekPort from the config and increment it by two. Use these two
        # ports to configure the XPUB/XSUB proxy component. The magic Port class
        # is located over in install.py
        base_port = ZeekControl.config.Config.get_option("ZeekPort")
        self.xpub_port = base_port
        self.xsub_port = base_port + 1
        ZeekControl.config.Config.set_option("ZeekPort", base_port + 2)

        # The manager address is used for listening of the XPUB/XSUB proxy.
        self.xpub_xsub_addr = ZeekControl.config.Config.manager().addr

        # If the address looks like an IPv6 address, put brackets around it
        # so that libzmq does not interpret it as a device and instead as
        # an IPv6 address.
        if ":" in self.xpub_xsub_addr:
            self.xpub_xsub_addr = "[" + self.xpub_xsub_addr + "]"

        # Check if this is a multi-node cluster (multiple IP addresses) and
        # tell the user about it.
        addrs = {n.addr for n in self.nodes()}

        # If use_curve_encryption is "auto", determine 0 or 1 based on
        # the number of different node addresses available. If it is
        # already "0" or "1", just make it the integer value.
        self.use_curve_encryption = (
            self.getOption("use_curve_encryption").lower().strip()
        )
        if len(addrs) > 1 and self.use_curve_encryption == "auto":
            self.use_curve_encryption = 1
        elif len(addrs) == 1 and self.use_curve_encryption == "auto":
            self.use_curve_encryption = 0
        elif self.use_curve_encryption in ["0", "1"]:
            self.use_curve_encryption = int(self.use_curve_encryption)
        elif self.use_curve_encryption in ["true", "false"]:
            self.use_curve_encryption = 1 if self.use_curve_encryption == "true" else 0
        else:
            raise ConfigurationError(
                f"invalid UseCurveEncryption value: {self.use_curve_encryption}"
            )

        # Store public and secret keys in in spool/zeromq/curve by default.
        curve_dir = self.getOption("curve_dir")
        self.server_publickey = Path(curve_dir) / "server_publickey"
        self.server_secretkey = Path(curve_dir) / "server_secretkey"
        self.client_publickey = Path(curve_dir) / "client_publickey"
        self.client_secretkey = Path(curve_dir) / "client_secretkey"

        # If encryption is enabled, create the spool directory for
        # the server and client keypairs and generate them if needed.
        if self.use_curve_encryption:
            os.makedirs(curve_dir, exist_ok=True)
            os.chmod(curve_dir, 0o700)

            if not self.server_publickey.exists() or not self.server_secretkey.exists():
                self.message("Generating ZeroMQ CURVE server keypair...")
                self.generate_keypair(self.server_publickey, self.server_secretkey)

            if not self.client_publickey.exists() or not self.client_secretkey.exists():
                self.message("Generating ZeroMQ CURVE client keypair...")
                self.generate_keypair(self.client_publickey, self.client_secretkey)

        if (
            not self.use_curve_encryption
            and len(addrs) > 1
            and not self.getOption("disable_unencrypted_warning")
        ):
            self.message(
                f'Warning: ZeroMQ encryption disabled, but multi-node cluster detected (IPs {", ".join(addrs)}).'
            )
            self.message(
                "\nYou may disable this warning by setting the following option in zeekctl.cfg:"
            )
            self.message(
                "\n    cluster_backend_zeromq.disable_unencrypted_warning = 1\n"
            )
            self.message(
                "\nYou may enable encryption by setting the following option.cfg:"
            )
            self.message(
                "\n    cluster_backend_zeromq.use_curve_encryption = 1 (or auto)\n"
            )

        # If any of the addresses used by nodes looks like an IPv6 address,
        # enable ZeroMQ IPv6 support via the configuration knob.
        self.ipv6 = False
        if any(":" in a for a in addrs):
            self.ipv6 = True

        return True

    def zeekctl_config(self):
        """
        Zeek 7.1 and later ship the following policy script to enable ZeroMQ.
        """
        script = "\n".join(
            [
                "# Enable ZeroMQ - the zeromq/connect script was deprecated with 8.1",
                '@if ( Version::at_least("8.1.0") )',
                "@load policy/frameworks/cluster/backend/zeromq",
                "@else",
                "@load policy/frameworks/cluster/backend/zeromq/connect",
                "@endif",
                "",
                f'redef Cluster::Backend::ZeroMQ::listen_xpub_endpoint = "tcp://{self.xpub_xsub_addr}:{self.xpub_port}";',
                f'redef Cluster::Backend::ZeroMQ::listen_xsub_endpoint = "tcp://{self.xpub_xsub_addr}:{self.xsub_port}";',
                f'redef Cluster::Backend::ZeroMQ::connect_xpub_endpoint = "tcp://{self.xpub_xsub_addr}:{self.xsub_port}";',
                f'redef Cluster::Backend::ZeroMQ::connect_xsub_endpoint = "tcp://{self.xpub_xsub_addr}:{self.xpub_port}";',
                "",
                f'redef Cluster::Backend::ZeroMQ::ipv6 = {"T" if self.ipv6 else "F"};',
                "",
            ]
        )

        # If CURVE encryption is enabled, redef the server and client
        # keys into the zeekctl-config file.
        if self.use_curve_encryption:

            def render_redef(path: Path) -> str:
                """
                Small helper to render a redef line given the path to
                the keypair files stored in curve_dir.
                """
                what = path.parts[-1]
                value = path.read_text().strip()
                if len(value) != 40:
                    raise ConfigurationError(f"CURVE key {what} not 40 bytes long")

                return "\n".join(
                    [
                        f"@if ( |Cluster::Backend::ZeroMQ::curve_{what}| == 0 )",
                        f'redef Cluster::Backend::ZeroMQ::curve_{what} = "{value}";',
                        "@endif",
                    ]
                )

            script += "\n".join(
                [
                    "",
                    "# Public and secret server keys for ZeroMQ CURVE encryption.",
                    render_redef(self.server_publickey),
                    render_redef(self.server_secretkey),
                    "",
                    "# Public and secret client keys for ZeroMQ CURVE encryption.",
                    render_redef(self.client_publickey),
                    render_redef(self.client_secretkey),
                    "",
                ]
            )

        # Usually this runs automatically on the manager, but Zeectl supports
        # standalone mode and the node doesn't know it should run the proxy
        # thread for WebSocket functionality.
        if ZeekControl.config.Config.standalone:
            script += "\n".join(
                [
                    "",
                    "# Standalone: Run the XPUB/XSUB thread in standalone mode",
                    "redef Cluster::Backend::ZeroMQ::run_proxy_thread = T;",
                    "",
                    "# Standalone: Subscribe to zeek.cluster.node.zeek.",
                    "#             for controllee WebSocket interactions.",
                    "event zeek_init()",
                    "	{",
                    '	Cluster::subscribe("zeek.cluster.node.zeek.");',
                    "	}",
                ]
            )

        return script



#
# A simple test plugin exercising most of a plugins capabilities. It just does
# some debugging output, nothing else.
#
# The plugin is off by default. To enable it, add "test.enabled=1" to zeekctl.cfg.

import ZeekControl.cmdresult
import ZeekControl.plugin


class TestPlugin(ZeekControl.plugin.Plugin):
    def __init__(self):
        super().__init__(apiversion=1)

    def name(self):
        return "TestPlugin"

    def prefix(self):
        return "test"

    def pluginVersion(self):
        return 1

    def init(self):
        if not self.getOption("enabled"):
            return False

        foo = self.getOption("foo")

        self.message("TestPlugin: Test initialized")
        self.message(f"TestPlugin: The value of foo is: {foo}")
        self.message(
            "TestPlugin: The current value of bar is: {}".format(self.getState("bar"))
        )

        for n in self.nodes():
            try:
                self.message(f"TestPlugin: mykey is: {n.test_mykey}")
            except AttributeError:
                self.message("TestPlugin: mykey is not set")

        for h in self.hosts():
            self.message(f"TestPlugin: host {h.host}")

        return True

    def options(self):
        return [
            ("foo", "int", 1, "Just a test option."),
            ("enabled", "bool", False, "Set to enable plugin"),
        ]

    def commands(self):
        return [("bar", "", "A test command from the Test plugin.")]

    def nodeKeys(self):
        return ["mykey"]

    def zeekctl_config(self):
        script = "# This is a test."
        return script

    def done(self):
        self.message("TestPlugin: done")

    def _nodes(self, nodes):
        if not nodes:
            return "<empty>"

        if isinstance(nodes[0], tuple):
            nodes = [n[0] for n in nodes]

        return ",".join([str(n) for n in nodes])

    def _results(self, results):
        if not results:
            return "<empty>"

        return ",".join([f"{str(n[0])}/{n[1]}" for n in results])

    def zeekProcessDied(self, node):
        self.message(f"TestPlugin: Zeek process died for {node}")

    def hostStatusChanged(self, host, status):
        self.message(f"TestPlugin: host status changed: {host} -> {status}")

    def cmd_custom(self, cmd, args, cmdout):
        results = ZeekControl.cmdresult.CmdResult()

        bar = self.getState("bar")
        if not bar:
            bar = 1

        self.setState("bar", bar + 1)
        self.message(f"TestPlugin: My command: {args}")

        return results

    def cmd_check_pre(self, nodes):
        self.message(f"TestPlugin: Test pre 'check':  {self._nodes(nodes)}")

    def cmd_check_post(self, results):
        self.message(f"TestPlugin: Test post 'check': {self._results(results)}")

    def cmd_nodes_pre(self):
        self.message("TestPlugin: Test pre 'nodes'")
        return True

    def cmd_nodes_post(self):
        self.message("TestPlugin: Test post 'nodes'")

    def cmd_config_pre(self):
        self.message("TestPlugin: Test pre 'config'")
        return True

    def cmd_config_post(self):
        self.message("TestPlugin: Test post 'config'")

    def cmd_deploy_pre(self):
        self.message("TestPlugin: Test pre 'deploy'")
        return True

    def cmd_deploy_post(self):
        self.message("TestPlugin: Test post 'deploy'")

    def cmd_exec_pre(self, cmdline):
        self.message(f"TestPlugin: Test pre 'exec':  {cmdline}")
        return True

    def cmd_exec_post(self, cmdline):
        self.message(f"TestPlugin: Test post 'exec': {cmdline}")

    def cmd_install_pre(self):
        self.message("TestPlugin: Test pre 'install'")
        return True

    def cmd_install_post(self):
        self.message("TestPlugin: Test post 'install'")

    def cmd_cron_pre(self, arg, watch):
        self.message(f"TestPlugin: Test pre 'cron':  {arg}/{watch}")
        return True

    def cmd_cron_post(self, arg, watch):
        self.message(f"TestPlugin: Test post 'cron': {arg}/{watch}")

    def cmd_restart_pre(self, nodes, clean):
        self.message(f"TestPlugin: Test pre 'restart':  {self._nodes(nodes)} ({clean})")

    def cmd_restart_post(self, nodes):
        self.message(f"TestPlugin: Test post 'restart': {self._nodes(nodes)}")

    def cmd_start_pre(self, nodes):
        self.message(f"TestPlugin: Test pre 'start':  {self._nodes(nodes)}")

    def cmd_start_post(self, results):
        self.message(f"TestPlugin: Test post 'start': {self._results(results)}")

    def cmd_stop_pre(self, nodes):
        self.message(f"TestPlugin: Test pre 'stop':  {self._nodes(nodes)}")

    def cmd_stop_post(self, results):
        self.message(f"TestPlugin: Test post 'stop': {self._results(results)}")

    def cmd_status_pre(self, nodes):
        self.message(f"TestPlugin: Test pre 'status':  {self._nodes(nodes)}")

    def cmd_status_post(self, nodes):
        self.message(f"TestPlugin: Test post 'status': {self._nodes(nodes)}")

    def cmd_update_pre(self, nodes):
        self.message(f"TestPlugin: Test pre 'update':  {self._nodes(nodes)}")

    def cmd_update_post(self, results):
        self.message(f"TestPlugin: Test post 'update': {self._results(results)}")

    def cmd_df_pre(self, nodes):
        self.message(f"TestPlugin: Test pre 'df':  {self._nodes(nodes)}")

    def cmd_df_post(self, nodes):
        self.message(f"TestPlugin: Test post 'df': {self._nodes(nodes)}")

    def cmd_diag_pre(self, nodes):
        self.message(f"TestPlugin: Test pre 'diag':  {self._nodes(nodes)}")

    def cmd_diag_post(self, nodes):
        self.message(f"TestPlugin: Test post 'diag': {self._nodes(nodes)}")

    def cmd_peerstatus_pre(self, nodes):
        self.message(f"TestPlugin: Test pre 'peerstatus':  {self._nodes(nodes)}")

    def cmd_peerstatus_post(self, nodes):
        self.message(f"TestPlugin: Test post 'peerstatus': {self._nodes(nodes)}")

    def cmd_netstats_pre(self, nodes):
        self.message(f"TestPlugin: Test pre 'netstats':  {self._nodes(nodes)}")

    def cmd_netstats_post(self, nodes):
        self.message(f"TestPlugin: Test post 'netstats': {self._nodes(nodes)}")

    def cmd_top_pre(self, nodes):
        self.message(f"TestPlugin: Test pre 'top':  {self._nodes(nodes)}")

    def cmd_top_post(self, nodes):
        self.message(f"TestPlugin: Test post 'top': {self._nodes(nodes)}")

    def cmd_cleanup_pre(self, nodes, all):
        self.message(f"TestPlugin: Test pre 'cleanup':  {self._nodes(nodes)} ({all})")

    def cmd_cleanup_post(self, nodes, all):
        self.message(f"TestPlugin: Test post 'cleanup': {self._nodes(nodes)} ({all})")

    def cmd_capstats_pre(self, nodes, interval):
        self.message(
            f"TestPlugin: Test pre 'capstats':  {self._nodes(nodes)} ({interval})"
        )

    def cmd_capstats_post(self, nodes, interval):
        self.message(
            f"TestPlugin: Test post 'capstats':  {self._nodes(nodes)} ({interval})"
        )

    def cmd_scripts_pre(self, nodes, check):
        self.message(f"TestPlugin: Test pre 'scripts':  {self._nodes(nodes)} ({check})")

    def cmd_scripts_post(self, nodes, check):
        self.message(f"TestPlugin: Test post 'scripts': {self._nodes(nodes)} ({check})")

    def cmd_print_pre(self, nodes, id):
        self.message(f"TestPlugin: Test pre 'print':  {self._nodes(nodes)} ({id})")

    def cmd_print_post(self, nodes, id):
        self.message(f"TestPlugin: Test post 'print': {self._nodes(nodes)} ({id})")

    def cmd_process_pre(self, trace, options, scripts):
        self.message(f"TestPlugin: Test pre 'process': {trace} {options} -- {scripts}")
        return True

    def cmd_process_post(self, trace, options, scripts, success):
        self.message(
            f"TestPlugin: Test post 'process': {trace} {options} -- {scripts} -> {success}"
        )



#
# Registry managing plugins.

import logging
import os
import sys

from ZeekControl import cmdresult, node, plugin

# Note, when changing this, also adapt doc string for Plugin.__init__.
_CurrentAPIVersion = 1


class PluginRegistry:
    def __init__(self):
        self._plugins = []
        self._dirs = []
        self._cmds = {}

    def _activeplugins(self):
        return filter(lambda p: p.activated, self._plugins)

    def addDir(self, dir):
        """Adds a directory to search for plugins."""
        if dir not in self._dirs:
            self._dirs += [dir]

    def loadPlugins(self, cmdout, executor):
        """Loads all plugins found in any of the added directories."""
        self._loadPlugins(cmdout)

        for p in self._plugins:
            p.executor = executor

    def initPluginOptions(self):
        """Initialize options for all loaded plugins."""
        for p in self._plugins:
            p._registerOptions()

    def initPlugins(self, cmdout):
        """Initialize all loaded plugins."""
        for p in self._plugins:
            p.activated = False

            try:
                init = p.init()
            except Exception as err:
                cmdout.warn(
                    f"Plugin '{p.name()}' not activated because its init() method raised exception: {err}"
                )
                continue

            if not init:
                logging.debug(
                    "Plugin '%s' not activated because its init() returned False",
                    p.name(),
                )
                continue

            p.activated = True

    def initPluginCmds(self):
        """Initialize commands provided by all activated plugins."""
        self._cmds = {}
        for p in self._activeplugins():
            for cmd, args, descr in p.commands():
                full_cmd = f"{p.prefix()}.{cmd}" if cmd else p.prefix()
                self._cmds[full_cmd] = (p, args, descr)

    def finishPlugins(self):
        """Shuts all plugins down."""
        for p in self._activeplugins():
            p.done()

    def hostStatusChanged(self, host, status):
        """Calls all plugins Plugin.hostStatusChanged_ methods; see there for
        parameter semantics."""
        for p in self._activeplugins():
            p.hostStatusChanged(host, status)

    def zeekProcessDied(self, node):
        """Calls all plugins Plugin.zeekProcessDied_ methods; see there for
        parameter semantics."""
        for p in self._activeplugins():
            p.zeekProcessDied(node)
            # TODO: Can we recognize when this is in use to warn about deprecation?
            p.broProcessDied(node)

    def cmdPreWithNodes(self, cmd, nodes, *args):
        """Executes the ``cmd_<XXX>_pre`` function for a command taking a list
        of nodes as its first argument. All other arguments are passed on.
        Returns the filtered node list as returned by the chain of all
        plugins."""

        method = f"cmd_{cmd}_pre"

        for p in self._activeplugins():
            func = getattr(p, method)
            new_nodes = func(nodes, *args)
            if new_nodes is not None:
                nodes = new_nodes

        return nodes

    def cmdPre(self, cmd, *args):
        """Executes the ``cmd_<XXX>_pre`` function for a command *not* taking
        a list of nodes as its first argument. All arguments are passed on.
        Returns True if no plugins returned False.
        """
        method = f"cmd_{cmd}_pre"
        result = True

        for p in self._activeplugins():
            func = getattr(p, method)
            if func(*args) == False:
                result = False

        return result

    def cmdPostWithNodes(self, cmd, nodes, *args):
        """Executes the ``cmd_<XXX>_post`` function for a command taking a list
        of nodes as its first argument. All other arguments are passed on.
        """
        method = f"cmd_{cmd}_post"

        for p in self._activeplugins():
            func = getattr(p, method)
            func(nodes, *args)

    def cmdPostWithResults(self, cmd, results, *args):
        """Executes the ``cmd_<XXX>_post`` function for a command taking a
        list of tuples ``(node, bool)`` as its first argument. All other
        arguments are passed on.
        """
        method = f"cmd_{cmd}_post"

        for p in self._activeplugins():
            func = getattr(p, method)
            func(results, *args)

    def cmdPost(self, cmd, *args):
        """Executes the ``cmd_<XXX>_post`` function for a command *not* taking
        a list of nodes as its first argument. All arguments are passed on.
        """
        method = f"cmd_{cmd}_post"

        for p in self._activeplugins():
            func = getattr(p, method)
            func(*args)

    def runCustomCommand(self, cmd, args, cmdout):
        """Runs a custom command *cmd* with string *args* as argument. Returns
        a CmdResult object which contains the command results."""

        try:
            myplugin, usage, descr = self._cmds[cmd]
        except LookupError:
            return cmdresult.CmdResult(ok=False, unknowncmd=True)

        prefix = myplugin.prefix()

        if cmd.startswith(f"{prefix}."):
            cmd = cmd[len(prefix) + 1 :]

        return myplugin.cmd_custom(cmd, args, cmdout)

    def getZeekctlConfig(self, cmdout):
        """Call the zeekctl_config method on all plugins in case a plugin
        needs to add some custom script code to the zeekctl-config.zeek file.
        Returns a string containing Zeek script code from the plugins.
        """

        extra_code = []

        for p in self._activeplugins():
            if p.broctl_config():
                cmdout.error(
                    f"Plugin '{p.name()}' uses discontinued method 'broctl_config'; use 'zeekctl_config' instead"
                )

            code = p.zeekctl_config()
            if code:
                # Make sure first character of returned string is a newline
                extra_code.append("")
                extra_code.append(f"# Begin code from {p.name()} plugin")
                extra_code.append(code)
                extra_code.append(f"# End code from {p.name()} plugin")

        if extra_code:
            # Make sure last character of returned string is a newline
            extra_code.append("")

        return "\n".join(extra_code)

    def allCustomCommands(self):
        """Returns a list of string tuples *(cmd, args, descr)* listing all
        commands defined by any plugin."""

        cmds = []

        for cmd in sorted(self._cmds.keys()):
            _, args, descr = self._cmds[cmd]
            cmds += [(cmd, args, descr)]

        return cmds

    def addNodeKeys(self):
        """Adds all plugins' node keys to the list of supported keys in
        ``Node``."""

        for p in self._plugins:
            for key in p.nodeKeys():
                key = f"{p.prefix()}_{key}"
                logging.debug("adding node key %s for plugin %s", key, p.name())
                node.Node.addKey(key)

    def _loadPlugins(self, cmdout):
        # Don't visit the same dir twice (this also prevents infinite
        # recursion when following symlinks).
        visited_dirs = set()

        for path in self._dirs:
            for root, dirs, files in os.walk(os.path.abspath(path), followlinks=True):
                stat = os.stat(root)
                visited_dirs.add((stat.st_dev, stat.st_ino))
                dirs_to_visit_next = []

                for dir in dirs:
                    stat = os.stat(os.path.join(root, dir))

                    if (stat.st_dev, stat.st_ino) not in visited_dirs:
                        dirs_to_visit_next.append(dir)

                dirs[:] = dirs_to_visit_next

                for name in sorted(files):
                    if name.endswith(".py") and not name.startswith("__"):
                        self._importPlugin(os.path.join(root, name[:-3]), cmdout)

    def _importPlugin(self, path, cmdout):
        sys.path = [os.path.dirname(path)] + sys.path

        try:
            module = __import__(os.path.basename(path))
        except Exception as e:
            cmdout.warn(f"cannot import plugin {path}: {e}")
            return
        finally:
            sys.path = sys.path[1:]

        found = False

        for cls in module.__dict__.values():
            try:
                is_plugin = issubclass(cls, plugin.Plugin)
            except TypeError:
                # cls is not a class.
                continue

            if is_plugin:
                found = True

                try:
                    p = cls()
                except Exception as e:
                    cmdout.warn(f"plugin class {cls.__name__} __init__ failed: {e}")
                    break

                # verify that the plugin overrides all required methods
                try:
                    logging.debug(
                        "Found plugin %s from %s (version %d, prefix %s)",
                        p.name(),
                        module.__file__,
                        p.pluginVersion(),
                        p.prefix(),
                    )
                except NotImplementedError:
                    cmdout.warn(
                        f"failed to load plugin at {path} because it doesn't override required methods"
                    )
                    continue

                if p.apiVersion() != _CurrentAPIVersion:
                    cmdout.warn(
                        f"failed to load plugin {p.name()} due to incompatible API version (uses {p.apiVersion():d}, but current is {_CurrentAPIVersion})"
                    )
                    continue

                if not p.prefix():
                    cmdout.warn(
                        f"failed to load plugin {p.name()} because prefix is empty"
                    )

                if "." in p.prefix() or " " in p.prefix():
                    cmdout.warn(
                        f"failed to load plugin {p.name()} because prefix contains dots or spaces"
                    )

                # Need to convert prefix to lowercase here, because a plugin
                # can override the prefix() method and might not return a
                # lowercase string.  Also, we don't allow two plugins to have
                # prefixes that differ only by case (due to the fact that
                # plugin option names include the prefix and are converted
                # to lowercase).
                pluginprefix = p.prefix().lower()

                sameprefix = False
                for i in self._plugins:
                    if pluginprefix == i.prefix().lower():
                        sameprefix = True
                        cmdout.warn(
                            f"failed to load plugin {p.name()} (prefix {p.prefix()}) due to plugin {i.name()} (prefix {i.prefix()}) having the same prefix"
                        )
                        break

                if not sameprefix:
                    self._plugins += [p]

        if not found:
            cmdout.warn(f"no plugin found in {module.__file__}")



import ZeekControl.config
import ZeekControl.plugin


class ClusterBackendBroker(ZeekControl.plugin.Plugin):
    def __init__(self):
        super().__init__(apiversion=1)

    def name(self):
        return "cluster_backend_broker"

    def pluginVersion(self):
        return 1

    def init(self):
        """
        Enable the plugin if ClusterBackend setting is Broker.
        """
        backend = self.getGlobalOption("ClusterBackend")
        if backend.lower() != "broker":
            return False

        # Switch topic separator for node topics to "/".
        if not self.getGlobalOption("ClusterTopicSeparator"):
            ZeekControl.config.Config.set_option("ClusterTopicSeparator", "/")

        return True

    def zeekctl_config(self):
        """
        If the version of Zeek already has the broker policy scripts, load
        these, otherwise Broker is anyhow enabled by default.
        """
        script = "\n".join(
            [
                "# Load Broker's policy scripts if they exist.",
                '@if ( can_load("policy/frameworks/cluster/backend/broker") )',
                "@load policy/frameworks/cluster/backend/broker",
                "@endif",
            ]
        )

        return script



"""Infrastructure for typed Zeek events."""

from typing import Any

from .brokertypes import (
    Boolean,
    DataType,
    Set,
    String,
    Type,
    Vector,
    ZeekEvent,
    from_py,
)
from .logs import LOG
from .types import (
    SerializableZeekType,
)


class Event(SerializableZeekType):
    NAME: str  # Name of the event, e.g. "Management::Controller::API::deploy_request"
    ARG_NAMES: list[str] = []  # Names of the arguments, e.g. "reqid"
    ARG_TYPES: list[type[DataType]] = []  # Types in Python, e.g. str

    def __init__(self, *args: Any) -> None:
        """Creates a Zeek event object.

        This expects the number of arguments contextualized above. The event
        name is not required since it's defined implicitly via the event class
        receiving the arguments.

        Raises:
            TypeError: when the given arguments, or number of arguments, don't
                match the expected ARG_TYPES or their number.
        """
        if len(args) != len(self.ARG_NAMES):
            raise TypeError(
                f"event argument length mismatch: have {len(args)}, expected {len(self.ARG_NAMES)}",
            )

        self.args = []

        for idx, arg in enumerate(args):
            # If the argument's type matches the required Broker type, we're done.
            if isinstance(arg, self.ARG_TYPES[idx]):
                self.args.append(arg)
                continue

            try:
                # When creating an event it can be convenient for the caller to
                # pass Python-native types. See if we can create brokertypes
                # types from them, to match the types actually specified when we
                # created the event classes.
                maybe_arg = from_py(arg)
            except TypeError as err:
                raise TypeError(
                    f"event argument type mismatch: argument "
                    f"{idx + 1} is {type(arg)}, {err}",
                ) from err

            # Again: if we now have a type match, we're done.
            if isinstance(maybe_arg, self.ARG_TYPES[idx]):
                self.args.append(maybe_arg)
                continue

            raise TypeError(
                f"event argument type mismatch: argument "
                f"{idx + 1} is {type(arg)}, should be {self.ARG_TYPES[idx]}",
            )

    def __getattr__(self, name: str) -> DataType:
        """Allow attribute-like access to event arguments."""
        try:
            idx = self.ARG_NAMES.index(name)
            return self.args[idx]
        except ValueError as err:
            raise AttributeError(
                f'event type {self.NAME} has no "{name}" argument',
            ) from err

    def __str__(self) -> str:
        # A list of pairs (argument name, typename)
        zeek_style_args = zip(
            self.ARG_NAMES, [str(type(arg)) for arg in self.args], strict=False
        )
        # That list, with each item now a string "<name>: <typename"
        zeek_style_arg_strings = [": ".join(arg) for arg in zeek_style_args]
        # A Zeek-looking event signature
        return self.NAME + "(" + ", ".join(zeek_style_arg_strings) + ")"

    def to_brokertype(self) -> ZeekEvent:
        return ZeekEvent(self.NAME, *self.args)

    @classmethod
    def from_brokertype(cls, data: Any) -> Any:
        # Verify that data is an event
        return Registry.make_event(data.name, data.args)


class Registry:
    """Functionality for event types and to instantiate typed events from data."""

    # Map from Zeek-level event names to Event classes. The make_event()
    # function uses this map to instantiate the right event class from
    # received Broker data.
    EVENT_TYPES: dict[str, type[Event]] = {}

    @staticmethod
    def make_event_class(
        name: str, arg_names: list[str], arg_types: list[type[Any]]
    ) -> type[Event]:
        """Factory function to generate a Zeek event class.

        Given an event name, event arguments, and corresponding argument types,
        the function generates a new Event class, registers it, and returns it.
        """
        res = type(name, (Event,), {})

        if len(arg_names) != len(arg_types):
            raise TypeError(
                f"error creating event type {name}: number of event "
                f"argument names and types must match ({len(arg_names)}/{len(arg_types)})",
            )

        for idx, typ in enumerate(arg_types):
            if not issubclass(typ, Type):
                raise TypeError(
                    f"event type creation error: argument {idx + 1}, "
                    f'"{arg_names[idx]}", is not a brokertype class',
                )
        res.NAME = name  # type: ignore
        res.ARG_NAMES = arg_names  # type: ignore
        res.ARG_TYPES = arg_types  # type: ignore

        # Register the new event type
        Registry.EVENT_TYPES[name] = res

        return res

    @staticmethod
    def make_event(name: str, *args: Any) -> Event | None:
        """This method allows constructing an Event instance from its name."""
        if name not in Registry.EVENT_TYPES:
            LOG.warning('received unexpected event "%s", skipping', name)
            return None

        LOG.debug('received event "%s"', name)
        return Registry.EVENT_TYPES[name](*args)


# Any Zeek object/record that's an event argument gets represented as a
# tuple below, reflecting Broker's representation thereof.

DeployRequest = Registry.make_event_class(
    "Management::Controller::API::deploy_request",
    ["reqid"],
    [String],
)

DeployResponse = Registry.make_event_class(
    "Management::Controller::API::deploy_response",
    ["reqid", "results"],
    [String, Vector],
)

GetConfigurationRequest = Registry.make_event_class(
    "Management::Controller::API::get_configuration_request",
    ["reqid", "deployed"],
    [String, Boolean],
)

GetConfigurationResponse = Registry.make_event_class(
    "Management::Controller::API::get_configuration_response",
    ["reqid", "result"],
    [String, Vector],
)

GetIdValueRequest = Registry.make_event_class(
    "Management::Controller::API::get_id_value_request",
    ["reqid", "id", "nodes"],
    [String, String, Set],
)

GetIdValueResponse = Registry.make_event_class(
    "Management::Controller::API::get_id_value_response",
    ["reqid", "results"],
    [String, Vector],
)

GetInstancesRequest = Registry.make_event_class(
    "Management::Controller::API::get_instances_request",
    ["reqid"],
    [String],
)

GetInstancesResponse = Registry.make_event_class(
    "Management::Controller::API::get_instances_response",
    ["reqid", "result"],
    [String, Vector],
)

GetNodesRequest = Registry.make_event_class(
    "Management::Controller::API::get_nodes_request",
    ["reqid"],
    [String],
)

GetNodesResponse = Registry.make_event_class(
    "Management::Controller::API::get_nodes_response",
    ["reqid", "results"],
    [String, Vector],
)

RestartRequest = Registry.make_event_class(
    "Management::Controller::API::restart_request",
    ["reqid", "nodes"],
    [String, Set],
)

RestartResponse = Registry.make_event_class(
    "Management::Controller::API::restart_response",
    ["reqid", "results"],
    [String, Vector],
)

StageConfigurationRequest = Registry.make_event_class(
    "Management::Controller::API::stage_configuration_request",
    ["reqid", "config"],
    [String, Vector],
)

StageConfigurationResponse = Registry.make_event_class(
    "Management::Controller::API::stage_configuration_response",
    ["reqid", "results"],
    [String, Vector],
)

TestNoopRequest = Registry.make_event_class(
    "Management::Controller::API::test_noop_request",
    ["reqid"],
    [String],
)

TestTimeoutRequest = Registry.make_event_class(
    "Management::Controller::API::test_timeout_request",
    ["reqid", "with_state"],
    [String, Boolean],
)

TestTimeoutResponse = Registry.make_event_class(
    "Management::Controller::API::test_timeout_response",
    ["reqid", "result"],
    [String, Vector],
)



import cmd

from ZeekControl.exceptions import CommandSyntaxError, InvalidNodeError, LockError


class ExitValueCmd(cmd.Cmd):
    def cmdloop(self, intro=None):
        """Repeatedly issue a prompt, accept input, parse an initial prefix
        off the received input, and dispatch to action methods, passing them
        the remainder of the line as argument.

        """

        self.preloop()
        if self.use_rawinput and self.completekey:
            try:
                import readline

                self.old_completer = readline.get_completer()
                readline.set_completer(self.complete)
                readline.parse_and_bind(self.completekey + ": complete")
            except ImportError:
                pass
        try:
            if intro is not None:
                self.intro = intro
            if self.intro:
                self.stdout.write(f"{self.intro}\n")
            self._stopping = False
            success = True
            while not self._stopping:
                if self.cmdqueue:
                    line = self.cmdqueue.pop(0)
                else:
                    if self.use_rawinput:
                        try:
                            line = input(self.prompt)
                        except EOFError:
                            line = "EOF"
                    else:
                        self.stdout.write(self.prompt)
                        self.stdout.flush()
                        line = self.stdin.readline()
                        if not line:
                            line = "EOF"
                        else:
                            line = line.rstrip("\r\n")
                line = self.precmd(line)
                try:
                    success = self.onecmd(line)
                except (CommandSyntaxError, InvalidNodeError, LockError) as err:
                    # Note that here we do not attempt to catch all ZeekControl
                    # exceptions; letting some just terminate the program to
                    # avoid getting in an unknown state (e.g. error while
                    # reloading the config).
                    success = False
                    print(f"Error: {err}")
                self.postcmd(False, line)
            self.postloop()
        finally:
            if self.use_rawinput and self.completekey:
                try:
                    import readline

                    readline.set_completer(self.old_completer)
                except ImportError:
                    pass
        return success



#
# Settings that specify the version of ZeekControl
#

VERSION = "2.6.0-56"
ZEEKBASE = "/usr/local"
CFGFILE = "/etc/zeek/zeekctl.cfg"
ZEEKSCRIPTDIR = "/usr/local/share/zeek"
LIBDIR = "/usr/local/lib"
LIBDIRINTERNAL = "/usr/local/lib/zeek/python"



# Store results of a zeekctl command.

from ZeekControl import node as node_mod


class CmdResult:
    """Class representing the result of a zeekctl command."""

    def __init__(self, ok=True, unknowncmd=False):
        # Command succeeded (True), or error occurred (False)
        self.ok = ok

        # If True, then the requested command does not exist.
        self.unknowncmd = unknowncmd

        # Number of Zeek nodes that command succeeded, and number that failed
        self.success_count = 0
        self.fail_count = 0

        # List of results for each node
        self.nodes = []

        # Results in the "nodes" list have been sorted (True), or not (False)
        self._sorted = False

        # List of arbitrary (key, value) pairs.
        self.keyval = []

    def to_dict(self):
        return {
            "success_count": self.success_count,
            "fail_count": self.fail_count,
            "nodes": self.get_node_data(),
        }

    def get_node_counts(self):
        """Return tuple (success, fail) of success and fail node counts."""

        return self.success_count, self.fail_count

    def get_node_data(self):
        """Return a list of tuples (node, success, data), where node is a
        Zeek node, success is a boolean, and data is a dictionary.
        """

        results = self.nodes

        if not self._sorted:
            results.sort(key=node_mod.sorttuple)
            self._sorted = True

        return results

    def get_node_output(self):
        """Return a list of tuples (node, success, output), where node is a
        Zeek node, success is a boolean, and output is a string.
        """

        results = []
        for node, success, out in self.nodes:
            output = out.get("_output", "")
            results.append((node, success, output))

        if not self._sorted:
            results.sort(key=node_mod.sorttuple)
            self._sorted = True

        return results

    def set_node_fail(self, node):
        """Records the fact that the given node failed."""

        self.nodes.append((node, False, {}))
        self.fail_count += 1
        self.ok = False

    def set_node_success(self, node):
        """Records the fact that the given node succeeded."""

        self.nodes.append((node, True, {}))
        self.success_count += 1

    def set_node_output(self, node, success, output):
        """Records the success status of the given node, and stores some
        output messages.  The node parameter is a Zeek node, success is a
        boolean, and output is a string.
        """

        self.nodes.append((node, success, {"_output": output}))
        if success:
            self.success_count += 1
        else:
            self.fail_count += 1
            self.ok = False

    def set_node_data(self, node, success, data):
        """Records the success status of the given node, and stores some data.
        The node parameter is a Zeek node, success is a boolean, and data is a
        dictionary.
        """

        self.nodes.append((node, success, data))
        if success:
            self.success_count += 1
        else:
            self.fail_count += 1
            self.ok = False
